@@ -134,6 +134,11 @@
   (blocks nil :type (or (simple-array (unsigned-byte 8) (*))
                         (simple-array (signed-byte 8) (*))
                         null))
+  ;; If all code locations in this function are in the same top level
+  ;; form, then this is the number of that form, otherwise NIL. If
+  ;; NIL, then each code location represented in the BLOCKS specifies
+  ;; the TLF number.
+  (tlf-number nil :type (or index null))
   ;; a vector describing the variables that the argument values are
   ;; stored in within this function. The locations are represented by
   ;; the ordinal number of the entry in the VARIABLES slot value. The
@@ -388,6 +393,10 @@
   ;; the universal time that the source was written, or NIL if
   ;; unavailable
   (created nil :type (or unsigned-byte null))
+  ;; The FILE-POSITIONs of the truly top level forms read from this
+  ;; file (if applicable). The vector element type will be chosen to
+  ;; hold the largest element.
+  (start-positions nil :type (or (simple-array * (*)) null) :read-only t)
   ;; Additional information from (WITH-COMPILATION-UNIT (:SOURCE-PLIST ...))
   (plist *source-plist* :read-only t))
 
@@ -410,18 +419,9 @@
   ;; Location contexts
   ;; A (simple-array * (*)) or a context if there's only one context.
   (contexts nil :type t :read-only t)
-  ;; Packed integers. Also can be a cons of that plus an alist which
-  ;; maps SB-C::COMPILED-DEBUG-FUN to SB-DI::COMPILED-DEBUG-FUN instances.
-  (tlf-num+offset nil :type (or integer cons)))
-
-;;; The TLF-NUMBER and CHAR-OFFSET of a compiled-debug-info can each be NIL,
-;;; but aren't often. However, to allow that, convert NIL to 0 and non-nil
-;;; value N to N+1.
-(defun pack-tlf-num+offset (tlf-number char-offset)
-  (with-adjustable-vector (v)
-    (write-var-integer (if tlf-number (1+ tlf-number) 0) v)
-    (write-var-integer (if char-offset (1+ char-offset) 0) v)
-    (integer-from-octets v)))
+  ;; An alist mapping SB-C::COMPILED-DEBUG-FUN
+  ;; to SB-DI::COMPILED-DEBUG-FUN instances.
+  (memo-cell nil :type list))
 
 ;;;; file reading
 ;;;;
@@ -439,17 +439,18 @@
              (:copier nil)
              (:print-object (lambda (s stream)
                               (print-unreadable-object (s stream :type t)
-                                (princ (file-info-name s) stream)))))
+                                (princ (file-info-truename s) stream)))))
   ;; If a file, the truename of the corresponding source file. If from
-  ;; a Lisp form, :LISP. If from a stream, :STREAM.
-  (name (missing-arg) :type (or pathname (eql :lisp)) :read-only t)
+  ;; a Lisp form, :LISP. In COMPILE-FILE, this gets filled lazily
+  ;; after the file gets opened.
+  (truename nil :type (or pathname null (eql :lisp)))
   ;; the external format that we'll call OPEN with, if NAME is a file.
   (external-format nil  :read-only t)
   ;; the defaulted, but not necessarily absolute file name (i.e. prior
   ;; to TRUENAME call.) Null if not a file. This is used to set
   ;; *COMPILE-FILE-PATHNAME*, and if absolute (a harmful constraint to be sure),
   ;; is dumped in the debug-info.
-  (untruename nil :type (or pathname null) :read-only t)
+  (pathname nil :type (or pathname null) :read-only t)
   ;; the file's write date (if relevant)
   (write-date nil :type (or unsigned-byte null)  :read-only t)
   ;; parallel vectors containing the forms read out of the file and
