@@ -53,18 +53,14 @@
     (labels ((retry-value (value)
                (or (typecase value
                      (fdefn (fdefn-fun value))
-                     (symbol
-                      (let ((fdefn (symbol-fdefn value)))
-                        (and fdefn
-                             (fdefn-fun fdefn))))
+                     (symbol (sb-impl::%symbol-function value))
                      (function value)
-                     (t
-                      (if (valid-function-name-p value)
-                          (let ((fdefn (find-fdefn value)))
-                            (and fdefn
-                                 (fdefn-fun fdefn)))
+                     ((satisfies legal-fun-name-p)
+                      (let ((fdefn (find-fdefn value)))
+                        (and fdefn (fdefn-fun fdefn))))
+                      (t
                           (still-bad "Bad value when restarting ~s: ~s"
-                                     name value))))
+                                     name value)))
                    (still-bad (if (fdefn-p value)
                                   "~S is still undefined"
                                   "Can't replace ~s with ~s because it is undefined")
@@ -336,10 +332,12 @@
                    (set-value value)))))
       (try condition))))
 
-(defun object-not-type-error (object type)
+(defun object-not-type-error (object type &optional (context nil context-p))
   (if (invalid-array-p object)
       (invalid-array-error object)
-      (let* ((context (sb-di:error-context))
+      (let* ((context (if context-p
+                          context
+                          (sb-di:error-context)))
              (condition
                (make-condition (if (and (%instancep object)
                                         (wrapper-invalid (%instance-wrapper object)))
@@ -430,6 +428,77 @@
   (bug "~@<failed AVER: ~2I~_~S~:>" form))
 (deferr unreachable-error ()
   (bug "Unreachable code reached"))
+
+(deferr mul-overflow-error (low high)
+  (destructuring-bind (raw-low raw-high) *current-internal-error-args*
+    (declare (ignorable raw-low raw-high))
+    (let ((type (or (sb-di:error-context)
+                    'fixnum)))
+      (object-not-type-error #+x86-64
+                             (* low high)
+                             #-x86-64
+                             (if (memq (sb-c:sc+offset-scn raw-low) `(,sb-vm:any-reg-sc-number
+                                                                      ,sb-vm:descriptor-reg-sc-number))
+                                 (ash (logior
+                                       (ash high sb-vm:n-word-bits)
+                                       (ldb (byte sb-vm:n-word-bits 0) (ash low sb-vm:n-fixnum-tag-bits)))
+                                      (- sb-vm:n-fixnum-tag-bits))
+                                 (logior
+                                  (ash high sb-vm:n-word-bits)
+                                  (ldb (byte sb-vm:n-word-bits 0) low)))
+                             type
+                             nil))))
+
+(sb-c::when-vop-existsp (:translate sb-c::unsigned+)
+  (flet ((err (x of cf)
+           (let* ((raw-x (car *current-internal-error-args*))
+                  (signed (= (sb-c:sc+offset-scn raw-x) sb-vm:signed-reg-sc-number)))
+             (let ((type (or (sb-di:error-context)
+                             'fixnum))
+                   (x (if signed
+                          (cond ((and of cf)
+                                 (dpb x (byte sb-vm:n-word-bits 0) -1))
+                                (of
+                                 (ldb (byte sb-vm:n-word-bits 0) x))
+                                (t
+                                 (bug "flags")))
+                          (cond (cf
+                                 (dpb 1 (byte 1 sb-vm:n-word-bits) x))
+                                (of
+                                 (sb-c::mask-signed-field sb-vm:n-word-bits x))
+                                (t
+                                 (dpb x (byte sb-vm:n-word-bits 0) -1))))))
+               (object-not-type-error x type nil)))))
+    (deferr add-sub-overflow-error (x)
+      (multiple-value-bind (of cf) (sb-vm::context-overflow-carry-flags *current-internal-error-context*)
+        (err x of cf)))
+
+   #+x86-64
+    (deferr sub-overflow-error (x)
+      (multiple-value-bind (of cf) (sb-vm::context-overflow-carry-flags *current-internal-error-context*)
+        (err x of (not cf)))))
+
+  (deferr signed-unsigned-add-overflow-error (x)
+    (let* ((type (or (sb-di:error-context)
+                     'fixnum))
+           (raw-x (car *current-internal-error-args*))
+           (signed (= (sb-c:sc+offset-scn raw-x) sb-vm:signed-reg-sc-number)))
+      (object-not-type-error (if signed
+                                 (if (logbitp (1- sb-vm:n-word-bits) x)
+                                     (ldb (byte sb-vm:n-word-bits 0) x)
+                                     (dpb 1 (byte 1 sb-vm:n-word-bits) x))
+                                 (multiple-value-bind (of cf)
+                                     (sb-vm::context-overflow-carry-flags *current-internal-error-context*)
+                                   (declare (ignore of))
+                                   (if cf
+                                       (sb-c::mask-signed-field sb-vm:n-word-bits x)
+                                       (dpb 1 (byte 1 sb-vm:n-word-bits) x))))
+                             type nil)))
+
+  (deferr sub-overflow2-error (x y)
+    (let ((type (or (sb-di:error-context)
+                    'fixnum)))
+      (object-not-type-error (- x y) type nil))))
 
 ;;;; INTERNAL-ERROR signal handler
 
@@ -462,7 +531,7 @@
              (if (and (>= error-number (length **internal-error-handlers**))
                       (< error-number (length sb-c:+backend-internal-errors+)))
                  (let ((context (sb-di:error-context)))
-                   (if (typep context '(cons (eql :struct-read)))
+                   (if (typep context '(cons (eql struct-read-context)))
                        ;; This was shoehorned into being a "type error"
                        ;; which isn't the best way to explain it to the user.
                        ;; However, from an API stance, it makes some sense to signal
