@@ -128,7 +128,9 @@
   (nevents (* dword)))
 
 (define-alien-routine "socket_input_available" int
-  (socket handle))
+  (socket handle)
+  (time long)
+  (utime long))
 
 (define-alien-routine "console_handle_p" boolean
   (handle handle))
@@ -139,17 +141,19 @@
 ;;; introspect it, so we have to try various things until we find
 ;;; something that works.  Returns true if there could be input
 ;;; available, or false if there is not.
-(defun handle-listen (handle)
+(defun handle-listen (handle &optional (time 0) (utime 0))
   (cond ((console-handle-p handle)
          (alien-funcall (extern-alien "win32_tty_listen"
                                       (function boolean handle))
                         handle))
         (t
          (with-alien ((avail dword))
-
-           (unless (zerop (peek-named-pipe handle nil 0 nil (addr avail) nil))
-             (return-from handle-listen (plusp avail))))
-         (let ((res (socket-input-available handle)))
+           (let* ((res (peek-named-pipe handle nil 0 nil (addr avail) nil))
+                  (code (get-last-error)))
+             (cond
+               ((not (zerop res)) (return-from handle-listen (plusp avail)))
+               ((= code error-broken-pipe) (return-from handle-listen t)))))
+         (let ((res (socket-input-available handle time utime)))
            (unless (zerop res)
              (return-from handle-listen (= res 1))))
          t)))
@@ -312,11 +316,13 @@
            (values result (if result 0 (get-last-error)))
            name nil))
 
+(defconstant +movefile-replace-existing+ 1)
+
 (defun sb-unix:unix-rename (name1 name2)
   (declare (type sb-unix:unix-pathname name1 name2))
-  (syscall (("MoveFile" t) lispbool system-string system-string)
+  (syscall (("MoveFileEx" t) lispbool system-string system-string dword)
            (values result (if result 0 (get-last-error)))
-           name1 name2))
+           name1 name2 +movefile-replace-existing+))
 
 (defun sb-unix::posix-getenv (name)
   (declare (type simple-string name))
@@ -343,7 +349,7 @@
 ;;;; Process time information
 
 (defconstant 100ns-per-internal-time-unit
-  (/ 10000000 sb-xc:internal-time-units-per-second))
+  (/ 10000000 internal-time-units-per-second))
 
 ;; FILETIME
 ;; The FILETIME structure is a 64-bit value representing the number of
@@ -376,6 +382,12 @@
                (addr ,kernel-time)
                (addr ,user-time))))
 
+#+64-bit
+(progn
+  (defun reinit-internal-real-time () nil)
+  (defun get-internal-real-time ()
+    (alien-funcall (extern-alien "get_monotonic_time" (function unsigned)))))
+#-64-bit
 (let ((epoch 0))
   (declare (unsigned-byte epoch))
   ;; FIXME: For optimization ideas see the unix implementation.
@@ -427,7 +439,6 @@
 (defconstant +filetime-unit+ 10000000)
 (defconstant +common-lisp-epoch-filetime-seconds+ 9435484800)
 
-#-sb-fluid
 (declaim (inline get-time-of-day))
 (defun get-time-of-day ()
   "Return the number of seconds and microseconds since the beginning of the
