@@ -13,65 +13,6 @@
 ;;;; files for more information.
 
 (in-package "SB-C")
-
-;;;; cost estimation
-
-;;; Return some sort of guess about the cost of a call to a function.
-;;; If the function has some templates, we return the cost of the
-;;; cheapest one, otherwise we return the cost of CALL-NAMED. Calling
-;;; this with functions that have transforms can result in relatively
-;;; meaningless results (exaggerated costs.)
-;;;
-;;; We special-case NULL, since it does have a source tranform and is
-;;; interesting to us.
-(defun fun-guessed-cost (name)
-  (declare (symbol name))
-  (let ((info (info :function :info name))
-        (call-cost (template-cost (template-or-lose 'call-named))))
-    (if info
-        (let ((templates (fun-info-templates info)))
-          (if templates
-              (template-cost (first templates))
-              (case name
-                (null (template-cost (template-or-lose 'if-eq)))
-                (t call-cost))))
-        call-cost)))
-
-;;; Return some sort of guess for the cost of doing a test against
-;;; TYPE. The result need not be precise as long as it isn't way out
-;;; in space. The units are based on the costs specified for various
-;;; templates in the VM definition.
-(defun type-test-cost (type)
-  (declare (type ctype type))
-  (or (when (eq type *universal-type*)
-        0)
-      (when (eq type *empty-type*)
-        0)
-      (let ((found (backend-type-predicate type)))
-        (if found
-            (+ (fun-guessed-cost found) (fun-guessed-cost 'eq))
-            nil))
-      (typecase type
-        (compound-type
-         (reduce #'+ (compound-type-types type) :key 'type-test-cost))
-        (member-type
-         (* (member-type-size type)
-            (fun-guessed-cost 'eq)))
-        (numeric-type
-         (* (if (numeric-type-complexp type) 2 1)
-            (fun-guessed-cost
-             (if (csubtypep type (specifier-type 'fixnum)) 'fixnump 'numberp))
-            (+ 1
-               (if (numeric-type-low type) 1 0)
-               (if (numeric-type-high type) 1 0))))
-        (cons-type
-         (+ (type-test-cost (specifier-type 'cons))
-            (fun-guessed-cost 'car)
-            (type-test-cost (cons-type-car-type type))
-            (fun-guessed-cost 'cdr)
-            (type-test-cost (cons-type-cdr-type type))))
-        (t
-         (fun-guessed-cost 'typep)))))
 
 (defun weaken-integer-type (type &key range-only)
   ;; FIXME: Our canonicalization isn't quite ideal for this. We get
@@ -127,7 +68,7 @@
     (weaken-integer-type-part type 'integer)))
 
 (defun-cached
-    (weaken-type :hash-bits 7 :hash-function #'type-hash-value)
+    (weaken-type :hash-bits 7 :hash-function #'sb-kernel::type-%bits)
     ((type eq))
   (declare (type ctype type))
   (cond ((named-type-p type)
@@ -143,33 +84,7 @@
          ;; Don't want to be putting CHARACTERs into BASE-STRINGs.
          (specifier-type 'base-char))
         (t
-         (let ((min-cost (type-test-cost type))
-               (min-type type)
-               (found-super nil))
-           (dolist (x *backend-type-predicates*)
-             (let* ((stype (car x))
-                    (samep (type= stype type)))
-               (when (or samep
-                         (and (csubtypep type stype)
-                              (not (union-type-p stype))))
-                 (let ((stype-cost (type-test-cost stype)))
-                   (when (or (< stype-cost min-cost)
-                             samep)
-                     ;; If the supertype is equal in cost to the type, we
-                     ;; prefer the supertype. This produces a closer
-                     ;; approximation of the right thing in the presence of
-                     ;; poor cost info.
-                     (setq found-super t
-                           min-type stype
-                           min-cost stype-cost))))))
-           ;; This used to return the *UNIVERSAL-TYPE* if no supertype was found,
-           ;; but that's too liberal: it's far too easy for the user to create
-           ;; a union type (which are excluded above), and then trick the compiler
-           ;; into trusting the union type... and finally ending up corrupting the
-           ;; heap once a bad object sneaks past the missing type check.
-           (if found-super
-               min-type
-               type)))))
+         type)))
 
 (defun weaken-values-type (type)
   (declare (type ctype type))
@@ -177,12 +92,9 @@
         ((not (values-type-p type))
          (weaken-type type))
         (t
-         (make-values-type :required (mapcar #'weaken-type
-                                             (values-type-required type))
-                           :optional (mapcar #'weaken-type
-                                             (values-type-optional type))
-                           :rest (acond ((values-type-rest type)
-                                         (weaken-type it)))))))
+         (make-values-type (mapcar #'weaken-type (values-type-required type))
+                           (mapcar #'weaken-type (values-type-optional type))
+                           (acond ((values-type-rest type) (weaken-type it)))))))
 
 ;;;; checking strategy determination
 
@@ -204,36 +116,22 @@
        (2 (weaken-values-type type))
        (3 type)))))
 
-(defun type-contains-fun-type-p (type)
-  (sb-kernel::map-type (lambda (type)
-                         (when (fun-type-p type)
-                           (return-from type-contains-fun-type-p t)))
-                       type))
-
 ;;; LVAR is an lvar we are doing a type check on and TYPES is a list
 ;;; of types that we are checking its values against. If we have
 ;;; proven that LVAR generates a fixed number of values, then for each
 ;;; value, we check whether it is cheaper to then difference between
 ;;; the proven type and the corresponding type in TYPES.
-(defun maybe-negate-check (lvar types original-types n-required)
+(defun lvar-types-to-check (lvar types original-types n-required)
   (declare (type lvar lvar) (list types original-types))
   (let ((ptypes (values-type-out (lvar-derived-type lvar) (length types))))
     (loop for p in ptypes
           and c in types
           and a in original-types
           and i from 0
-          for cc = (if (>= i n-required)
-                       (type-union c (specifier-type 'null))
-                       c)
-          for diff = (type-difference p cc)
-          collect (if (and diff
-                           ;; FUN-TYPE is weakend to FUNCTION, can't
-                           ;; use that for invereted type testes
-                           (not (type-contains-fun-type-p diff))
-                           (< (type-test-cost diff)
-                              (type-test-cost cc)))
-                      (list t diff a)
-                      (list nil cc a)))))
+          collect (list (if (>= i n-required)
+                            (type-union c (specifier-type 'null))
+                            c)
+                        a))))
 
 ;;; Determine if CAST can be checked.
 ;;; We may check only fixed number of values; in any case the number
@@ -279,14 +177,14 @@
                 (not (values-type-rest dtype)))
            ;; we [almost] know how many values are produced
            (values :simple
-                   (maybe-negate-check value
-                                       (values-type-out ctype n-required)
-                                       (values-type-out atype n-required)
-                                       n-required)))
+                   (lvar-types-to-check value
+                                        (values-type-out ctype n-required)
+                                        (values-type-out atype n-required)
+                                        n-required)))
           ((lvar-single-value-p lvar)
            ;; exactly one value is consumed
            (principal-lvar-single-valuify lvar)
-           (values :simple (maybe-negate-check value
+           (values :simple (lvar-types-to-check value
                                                (list (single-value-type ctype))
                                                (list (single-value-type atype))
                                                n-required)))
@@ -294,7 +192,7 @@
                 (eq (mv-combination-kind dest) :local)
                 (singleton-p (mv-combination-args dest)))
            ;; we know the number of consumed values
-           (values :simple (maybe-negate-check value
+           (values :simple (lvar-types-to-check value
                                                (adjust-list (values-type-types ctype)
                                                             n-consumed
                                                             *universal-type*)
@@ -312,6 +210,9 @@
   (let* ((lvar (node-lvar cast))
          (dest (and lvar (lvar-dest lvar))))
     (and (combination-p dest)
+         (or (not (combination-fun-info dest))
+             ;; fixed-args functions do not check their arguments.
+             (not (ir1-attributep (fun-info-attributes (combination-fun-info dest)) fixed-args)))
          ;; The theory is that the type assertion is from a declaration on the
          ;; callee, so the callee should be able to do the check. We want to
          ;; let the callee do the check, because it is possible that by the
@@ -334,8 +235,19 @@
          ;; deemed unreachable?
          (and
           (almost-immediately-used-p lvar cast)
-          (values (values-subtypep (lvar-externally-checkable-type lvar)
-                                   (cast-type-to-check cast)))))))
+          (if (and (lvar-fun-is (combination-fun dest)
+                                '(hairy-data-vector-set/check-bounds
+                                  hairy-data-vector-ref/check-bounds
+                                  hairy-data-vector-ref
+                                  hairy-data-vector-set))
+                   (eql (car (combination-args dest))
+                        lvar))
+              ;; These functions work on all arrays, but the error
+              ;; message is about vectors, which is used more frequently.
+              (csubtypep (specifier-type 'vector)
+                         (single-value-type (cast-type-to-check cast)))
+              (values-subtypep (lvar-externally-checkable-type lvar)
+                               (cast-type-to-check cast)))))))
 
 ;; Type specifiers handled by the general-purpose MAKE-TYPE-CHECK-FORM are often
 ;; trivial enough to have an internal error number assigned to them that can be
@@ -456,29 +368,22 @@
   (let* ((temps (make-gensym-list (length types)))
          (context (cast-context cast))
          (restart (and (eq context :restart)
-                       (setf context (opaquely-quote (make-restart-location))))))
+                       (setf context (make-restart-location)))))
     (lambda (dummy)
       `(multiple-value-bind ,temps ,dummy
          ,@(mapcar
             (lambda (temp %type)
-              (destructuring-bind (not type-to-check
-                                   type-to-report) %type
-                (let* ((spec
-                         (let ((*unparse-fun-type-simplify* t))
-                           (type-specifier type-to-check)))
-                       (test (if not `(not ,spec) spec)))
-                  `(progn
-                     (unless
-                         ,(with-ir1-environment-from-node cast ;; it performs its own inlining of SATISFIES
-                            (%source-transform-typep temp test))
-                       ,(internal-type-error-call temp
-                                                  (if (fun-designator-type-p type-to-report)
-                                                      ;; Simplify
-                                                      (specifier-type 'function-designator)
-                                                      type-to-report)
-                                                  context))
-                     ,@(and restart
-                            `((restart-point ,restart)))))))
+              (destructuring-bind (type-to-check type-to-report) %type
+                `(progn
+                   (unless (typep ,temp ',(type-specifier type-to-check t))
+                     ,(internal-type-error-call temp
+                                                (if (fun-designator-type-p type-to-report)
+                                                    ;; Simplify
+                                                    (specifier-type 'function-designator)
+                                                    type-to-report)
+                                                context))
+                   ,@(and restart
+                          `((restart-point ,restart))))))
             temps
             types)
          (values ,@temps)))))
@@ -501,41 +406,35 @@
   (let* ((lvar (node-lvar cast))
          (dest (and lvar (lvar-dest lvar)))
          (value (cast-value cast))
-         (atype (cast-asserted-type cast))
-         (condition 'type-warning)
-         (not-ok-uses '()))
+         (atype (cast-asserted-type cast)))
     (do-uses (use value)
       (let ((dtype (node-derived-type use)))
-        (cond ((values-types-equal-or-intersect dtype atype)
-               (setf condition 'type-style-warning))
-              ((cast-mismatch-from-inlined-p cast use))
-              (t
-               (push use not-ok-uses)))))
-    (dolist (use (nreverse not-ok-uses))
-      (let* ((*compiler-error-context* use)
-             (dtype (node-derived-type use))
-             (what (when (and (combination-p dest)
-                              (eq (combination-kind dest) :local))
-                     (let ((lambda (combination-lambda dest))
-                           (pos (position-or-lose
-                                 lvar (combination-args dest))))
-                       (format nil "~:[A possible~;The~] binding of ~S"
-                               (and (lvar-has-single-use-p lvar)
-                                    (eq (functional-kind lambda) :let))
-                               (leaf-source-name (elt (lambda-vars lambda)
-                                                      pos)))))))
-        (cond ((and (ref-p use) (constant-p (ref-leaf use)))
-               (warn condition
-                     :format-control "~:[This~;~:*~A~] is not a ~
+        (unless (or (values-types-equal-or-intersect dtype atype)
+                    (cast-mismatch-from-inlined-p cast use))
+          (let* ((*compiler-error-context* use)
+                 (dtype (node-derived-type use))
+                 (what (when (and (combination-p dest)
+                                  (eq (combination-kind dest) :local))
+                         (let ((lambda (combination-lambda dest))
+                               (pos (position-or-lose
+                                     lvar (combination-args dest))))
+                           (format nil "~:[A possible~;The~] binding of ~S"
+                                   (and (lvar-has-single-use-p lvar)
+                                        (eq (functional-kind lambda) :let))
+                                   (leaf-source-name (elt (lambda-vars lambda)
+                                                          pos)))))))
+            (cond ((and (ref-p use) (constant-p (ref-leaf use)))
+                   (warn 'type-style-warning
+                         :format-control "~:[This~;~:*~A~] is not a ~
                        ~<~%~9T~:;~/sb-impl:print-type/:~>~% ~S"
-                     :format-arguments
-                     (list what atype (constant-value (ref-leaf use)))))
-              (t
-               (warn condition
-                     :format-control
-                      "~:[Result~;~:*~A~] is a ~/sb-impl:print-type/, ~
+                         :format-arguments
+                         (list what atype (constant-value (ref-leaf use)))))
+                  (t
+                   (warn 'type-style-warning
+                         :format-control
+                         "~:[Result~;~:*~A~] is a ~/sb-impl:print-type/, ~
                        ~<~%~9T~:;not a ~/sb-impl:print-type/.~>"
-                     :format-arguments (list what dtype atype)))))))
+                         :format-arguments (list what dtype atype)))))))))
   (values))
 
 ;;; Loop over all blocks in COMPONENT that have TYPE-CHECK set,
@@ -605,14 +504,14 @@
                            (format nil
                                    "~%It allows an unknown number of values, consider using~@
                                     ~/sb-impl:print-type/."
-                                   (make-values-type :required (values-type-required value-type)
-                                                     :optional (values-type-optional value-type))))
+                                   (make-values-type (values-type-required value-type)
+                                                     (values-type-optional value-type))))
                           ((values-type-optional value-type)
                            (format nil
                                    "~%It allows a variable number of values, consider using~@
                                     ~/sb-impl:print-type/."
-                                   (make-values-type :required (append (values-type-required value-type)
-                                                                       (values-type-optional value-type)))))
+                                   (make-values-type (append (values-type-required value-type)
+                                                             (values-type-optional value-type)))))
                           ("")))))
                (setf (cast-type-to-check cast) *wild-type*)
                (setf (cast-%type-check cast) nil)))))))
