@@ -274,10 +274,14 @@
   (integer integer) rational
   (no-verify-arg-count unsafely-flushable))
 
-(defknown (two-arg-< two-arg-= two-arg->)
+(defknown (two-arg-< two-arg-= two-arg-> two-arg-<= two-arg->=)
     (number number) boolean
     (no-verify-arg-count))
 
+(defknown (range< range<= range<<= range<=<)
+    (fixnum real fixnum) boolean
+    (foldable flushable movable no-verify-arg-count))
+
 (defknown (two-arg-gcd two-arg-lcm two-arg-and two-arg-ior two-arg-xor two-arg-eqv)
   (integer integer) integer
   (no-verify-arg-count))
@@ -335,7 +339,10 @@
 (defknown (rationalize) (real) rational
   (movable foldable flushable recursive))
 
-(defknown (numerator denominator) (rational) integer
+(defknown numerator (rational) integer
+  (movable foldable flushable))
+
+(defknown denominator (rational) (integer 1)
   (movable foldable flushable))
 
 (defknown (floor ceiling round)
@@ -349,8 +356,11 @@
 (defknown unary-truncate (real) (values integer real)
   (movable foldable flushable no-verify-arg-count))
 
+;;; KLUDGE: Don't fold these, the compiler may call it on floats that
+;;; do not truncate into a bignum, and the functions do not check
+;;; their input values and produce corrupted memory.
 (defknown unary-truncate-single-float-to-bignum (single-float) (values bignum (eql $0f0))
-   (foldable movable flushable fixed-args))
+    (#+(or) foldable movable flushable fixed-args))
 (defknown unary-truncate-double-float-to-bignum (double-float)
     (values #+64-bit bignum #-64-bit integer
             (and
@@ -358,12 +368,19 @@
                     (not (or riscv ppc64))) ;; they can't survive cold-init
              (eql $0d0)
              double-float))
-   (foldable movable flushable fixed-args))
+   (#+(or) foldable movable flushable fixed-args))
 
 (defknown %unary-truncate-single-float-to-bignum (single-float) bignum
-   (foldable movable flushable fixed-args))
+   (#+(or) foldable movable flushable fixed-args))
 (defknown %unary-truncate-double-float-to-bignum (double-float) bignum
-   (foldable movable flushable fixed-args))
+   (#+(or) foldable movable flushable fixed-args))
+
+(defknown (subtract-bignum add-bignums) (bignum bignum) integer
+    (movable flushable no-verify-arg-count))
+(defknown (add-bignum-fixnum subtract-bignum-fixnum) (bignum fixnum) integer
+    (movable flushable no-verify-arg-count))
+(defknown subtract-fixnum-bignum (fixnum bignum) integer
+    (movable flushable no-verify-arg-count))
 
 (defknown sxhash-bignum-double-float (double-float) hash-code
   (foldable movable flushable fixed-args))
@@ -2175,7 +2192,7 @@
   (movable foldable flushable))
 
 ;; FIXME: should T be be (OR INSTANCE FUNCALLABLE-INSTANCE) etc?
-(defknown slot-value (t symbol) t (any))
+(defknown (slot-value slot-makunbound) (t symbol) t (any))
 (defknown (slot-boundp slot-exists-p) (t symbol) boolean)
 (defknown sb-pcl::set-slot-value (t symbol t) t (any))
 
@@ -2216,19 +2233,24 @@
 (defknown fixnum*
   (fixnum fixnum t)
   fixnum
-  (movable foldable always-translatable))
+  (movable always-translatable))
 
 (defknown (signed* signed+ signed-)
   (sb-vm:signed-word sb-vm:signed-word t)
   sb-vm:signed-word
-  (movable foldable commutative always-translatable))
+  (movable always-translatable))
 
 (defknown (unsigned* unsigned+ unsigned-)
   (word word t)
   word
-  (movable foldable commutative always-translatable))
+  (movable always-translatable))
 
 (defknown (unsigned+signed unsigned-signed)
   (word sb-vm:signed-word t)
-  word
-  (movable foldable commutative always-translatable))
+  integer
+  (movable always-translatable))
+
+(defknown (signed-unsigned)
+  (sb-vm:signed-word word t)
+  integer
+  (movable always-translatable))
