@@ -252,6 +252,7 @@ static inline lispobj compute_lispobj(lispobj* base_addr) {
   ;; Address of the one mandatory 'struct arena_memblk' for this arena
   (first-block 0 :type word)
   ;; Arena allocation parameters
+  (original-size 0 :type word)
   (growth-amount 0 :type word) ; additive
   (max-extensions 0 :type word)
   ;; Sum of sizes of currently allocated blocks
@@ -264,6 +265,8 @@ static inline lispobj compute_lispobj(lispobj* base_addr) {
   (extension-count 0 :type word)
   ;; Small integer identifier starting from 0
   (index 0 :type fixnum)
+  ;; T if all memory has been protected with PROT_NONE (for debugging)
+  hidden
   ;; a counter that increments on each rewind, and which can be used by a threads
   ;; in a pool to detect that their cached TLAB pointers are invalid
   (token 0 :type word)
