@@ -117,26 +117,15 @@ RETURN-FROM can be used to exit the form."
                                   :cleanup cleanup)))
       (ir1-convert-progn-body dummy next result forms))))
 
+;;; We make NEXT start a block just so that it will have a block
+;;; assigned. People assume that when they pass a ctran into
+;;; IR1-CONVERT as NEXT, it will have a block when it is done.
 (def-ir1-translator return-from ((name &optional value) start next result)
   "RETURN-FROM name value
 
 Evaluate the VALUE, returning its values from the lexically enclosing
 block NAME. This is constrained to be used only within the dynamic
 extent of the block."
-  ;; old comment:
-  ;;   We make NEXT start a block just so that it will have a block
-  ;;   assigned. People assume that when they pass a ctran into
-  ;;   IR1-CONVERT as NEXT, it will have a block when it is done.
-  ;; KLUDGE: Note that this block is basically fictitious. In the code
-  ;;   (BLOCK B (RETURN-FROM B) (SETQ X 3))
-  ;; it's the block which answers the question "which block is
-  ;; the (SETQ X 3) in?" when the right answer is that (SETQ X 3) is
-  ;; dead code and so doesn't really have a block at all. The existence
-  ;; of this block, and that way that it doesn't explicitly say
-  ;; "I'm actually nowhere at all" makes some logic (e.g.
-  ;; BLOCK-HOME-LAMBDA-OR-NULL) more obscure, and it might be better
-  ;; to get rid of it, perhaps using a special placeholder value
-  ;; to indicate the orphanedness of the code.
   (ctran-starts-block next)
   (let* ((found (or (lexenv-find name blocks)
                     (compiler-error "return for unknown block: ~S" name)))
@@ -324,25 +313,27 @@ Evaluate the FORMS in the specified SITUATIONS (any of :COMPILE-TOPLEVEL,
           (unless (listp arglist)
             (fail "The local macro argument list ~S is not a list."
                   arglist))
-          `(,name macro .
-                  ;; I guess the reason we want to compile here rather than possibly
-                  ;; using an interpreted lambda is that we generate the usual gamut
-                  ;; of style-warnings and such. One might wonder if this could somehow
-                  ;; go through the front-most part of the front-end, to deal with
-                  ;; semantics, but then generate an interpreted function or something
-                  ;; more quick to emit than machine code.
-                  ,(compile-in-lexenv
-                    (let (#-sb-xc-host
-                          (*macro-policy*
-                            ;; Make sure to save the sources if an
-                            ;; inlined functions closes over this
-                            ;; macro.
-                            (process-optimize-decl
-                             '(optimize (store-source-form 3))
-                             *macro-policy*)))
-                      (make-macro-lambda nil arglist body 'macrolet name))
-                    lexenv
-                    nil nil nil t nil))))))) ; name source-info tlf ephemeral errorp
+          (let (source-info
+                tlf)
+            #-sb-xc-host
+            (when (boundp '*source-paths*)
+              (push `(declare (source-form ,definition)) body)
+              (setf source-info *source-info*
+                    tlf (let ((path (or (get-source-path definition)
+                                        (and (boundp '*current-path*)
+                                             *current-path*))))
+                          (source-path-tlf-number path))))
+            `(,name macro .
+                    ;; I guess the reason we want to compile here rather than possibly
+                    ;; using an interpreted lambda is that we generate the usual gamut
+                    ;; of style-warnings and such. One might wonder if this could somehow
+                    ;; go through the front-most part of the front-end, to deal with
+                    ;; semantics, but then generate an interpreted function or something
+                    ;; more quick to emit than machine code.
+                    ,(compile-in-lexenv
+                      (make-macro-lambda nil arglist body 'macrolet name)
+                      lexenv
+                      nil source-info tlf t nil))))))))
 
 (defun funcall-in-macrolet-lexenv (definitions fun context)
   (%funcall-in-foomacrolet-lexenv
@@ -461,11 +452,40 @@ body, references to a NAME will effectively be replaced with the EXPANSION."
       (bug "%PRIMITIVE was used with an unknown values template."))
 
     (ir1-convert start next result
-                 `(%%primitive ',name
+                 `(%%primitive ',template
                                ',(eval-info-args
                                   (subseq args required min))
                                ,@(subseq args 0 required)
                                ,@(subseq args min)))))
+
+(defmacro inline-%primitive (template &rest args)
+  (let* ((required (length (template-arg-types template)))
+         (info (template-info-arg-count template))
+         (min (+ required info))
+         (nargs (length args)))
+    (if (template-more-args-type template)
+        (when (< nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants at least ~R."
+               nargs
+               min))
+        (unless (= nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants exactly ~R."
+               nargs
+               min)))
+
+    (when (template-conditional-p template)
+      (bug "%PRIMITIVE was used with a conditional template."))
+
+    (when (template-more-results-type template)
+      (bug "%PRIMITIVE was used with an unknown values template."))
+
+    `(%%primitive ',template
+                  ',(eval-info-args
+                     (subseq args required min))
+                  ,@(subseq args 0 required)
+                  ,@(subseq args min))))
 
 ;;;; QUOTE
 
@@ -513,6 +533,7 @@ Return VALUE without evaluating it."
 ;;; un-merged pathnames. I'm not daring enough to change it for everyone.
 ;;; It defaults to what it should, and is changed before saving the image.
 ;;;
+;;; FIXME: can't we just get rid of this and _never_ use TRUENAME?
 (declaim (type (member pathname truename) *name-context-file-path-selector*))
 (defglobal *name-context-file-path-selector* 'pathname)
 
@@ -540,6 +561,29 @@ Return VALUE without evaluating it."
       (list :in context))))
 
 ;;;; FUNCTION and NAMED-LAMBDA
+
+;;; `(NAMED-LAMBDA ,NAME ,@REST) is like `(FUNCTION (LAMBDA ,@REST)),
+;;; except that the value of NAME is passed to the compiler for use in
+;;; creation of debug information for the resulting function.
+;;;
+;;; NAME can be a legal function name or some arbitrary other thing.
+;;;
+;;; If NAME is a legal function name, then the caller should be
+;;; planning to set (FDEFINITION NAME) to the created function.
+;;; (Otherwise the debug names will be inconsistent and thus
+;;; unnecessarily confusing.)
+;;;
+;;; Arbitrary other things are appropriate for naming things which are
+;;; not the FDEFINITION of NAME. E.g.
+;;;   NAME = (:FLET FOO BAR)
+;;; for the FLET function in
+;;;   (DEFUN BAR (X)
+;;;     (FLET ((FOO (Y) (+ X Y)))
+;;;       FOO))
+;;; or
+;;;   NAME = (:METHOD PRINT-OBJECT :AROUND (STARSHIP T))
+;;; for the function used to implement
+;;;   (DEFMETHOD PRINT-OBJECT :AROUND ((SS STARSHIP) STREAM) ...).
 (defun name-lambdalike (thing)
   (case (car thing)
     ((named-lambda)
@@ -547,9 +591,6 @@ Return VALUE without evaluating it."
          `(lambda ,(strip-lambda-list (third thing) :name) ,(name-context))))
     ((lambda)
      `(lambda ,(strip-lambda-list (second thing) :name) ,@(name-context)))
-    ((lambda-with-lexenv)
-     ;; FIXME: Get the original DEFUN name here.
-     `(lambda ,(fifth thing)))
     (otherwise
      (compiler-error "Not a valid lambda expression:~%  ~S"
                      thing))))
@@ -566,7 +607,7 @@ Return VALUE without evaluating it."
 ;;; apparent function associated to it.
 (defun find-or-convert-fun-leaf (thing start)
   (cond
-    ((typep thing '(cons (member lambda named-lambda lambda-with-lexenv)))
+    ((typep thing '(cons (member lambda named-lambda)))
      (let ((ctran (make-ctran))
            (leaf (ir1-convert-lambdalike thing
                                          :debug-name (name-lambdalike thing))))
@@ -687,6 +728,12 @@ be a lambda expression."
 (def-ir1-translator %funcall-lvar ((function &rest args) start next result)
   (ir1-convert-combination-args function start next result args))
 
+(def-ir1-translator %funcall-no-nargs ((function &rest args) start next result)
+  (let ((ctran (make-ctran))
+        (fun-lvar (make-lvar)))
+    (ir1-convert start ctran fun-lvar `(the function ,function))
+    (ir1-convert-combination-args fun-lvar ctran next result args :pass-nargs nil)))
+
 ;;; This source transform exists to reduce the amount of work for the
 ;;; compiler. If the called function is a FUNCTION form, then convert
 ;;; directly to %FUNCALL, instead of waiting around for type
@@ -789,7 +836,8 @@ have been evaluated."
                                    :debug-name (debug-name 'let bindings))))
                          (reference-leaf start ctran fun-lvar fun))
                        (values next result))))
-           (ir1-convert-combination-args fun-lvar ctran next result values)))))
+           (ir1-convert-combination-args fun-lvar ctran next result values
+                                         :arg-source-forms bindings)))))
 
 (def-ir1-translator let* ((bindings &body body)
                           start next result)
@@ -800,7 +848,8 @@ form to reference any of the previous VARS."
   (multiple-value-bind (vars values forms decls) (parse-letish bindings body 'let*)
     (processing-decls (decls vars nil next result post-binding-lexenv)
       (ir1-convert-aux-bindings
-       start next result forms vars values post-binding-lexenv))))
+       start next result forms vars values post-binding-lexenv
+       :value-source-forms bindings))))
 
 ;;; logic shared between IR1 translators for LOCALLY, MACROLET,
 ;;; and SYMBOL-MACROLET
@@ -1171,13 +1220,14 @@ care."
       (let* ((name (first things))
              (value-form (second things))
              (leaf (or (lexenv-find name vars) (find-free-var name))))
+        (maybe-note-undefined-variable-reference leaf name)
         (etypecase leaf
           (leaf
            (when (constant-p leaf)
              (compiler-error "~S is a constant and thus can't be set." name))
            (when (lambda-var-p leaf)
              (let ((home-lambda (ctran-home-lambda-or-null start)))
-               (when home-lambda
+               (when (and home-lambda (neq (lambda-var-home leaf) home-lambda))
                  (sset-adjoin leaf (lambda-calls-or-closes home-lambda))))
              (when (lambda-var-ignorep leaf)
                ;; ANSI's definition of "Declaration IGNORE, IGNORABLE"
