@@ -105,11 +105,11 @@
 (defvar *current-path*)
 
 (defun call-with-current-source-form (thunk &rest forms)
-  (let ((*current-path* (when (and (some #'identity forms)
-                                   (boundp '*source-paths*))
-                          (or (some #'get-source-path forms)
-                              (when (boundp '*current-path*)
-                                *current-path*)))))
+  (let ((*current-path* (or (and (some #'identity forms)
+                                 (boundp '*source-paths*)
+                                 (some #'get-source-path forms))
+                            (and (boundp '*current-path*)
+                                 *current-path*))))
     (funcall thunk)))
 
 (defvar *derive-function-types* nil
@@ -164,58 +164,17 @@
               ;; FIXME: this seems to omit FUNCTIONAL
               (when (defined-fun-p fun)
                 (return-from fun-lexically-notinline-p
-                  (eq (defined-fun-inlinep fun) 'notinline))))))))
+                  (eq (defined-fun-inlinep fun) 'notinline)))
+              (loop for data in (lexenv-user-data env)
+                    when (and (eq (car data) 'no-compiler-macro)
+                              (eq (cdr data) name))
+                    do
+                    (return-from fun-lexically-notinline-p
+                      t)))))))
     ;; If ANSWER is NIL, go for the global value
     (eq (or answer (info :function :inlinep name)) 'notinline)))
 
 
-;;;; code coverage
-
-;;; Check the policy for whether we should generate code coverage
-;;; instrumentation. If not, just return the original START
-;;; ctran. Otherwise insert code coverage instrumentation after
-;;; START, and return the new ctran.
-(defun instrument-coverage (start mode form
-                            &aux (metadata (coverage-metadata *compilation*)))
-  ;; We don't actually use FORM for anything, it's just convenient to
-  ;; have around when debugging the instrumentation.
-  (declare (ignore form))
-  (if (and metadata
-           (policy *lexenv* (> store-coverage-data 0))
-           *allow-instrumenting*)
-      (let ((path (source-path-original-source *current-path*)))
-        (when mode
-          (push mode path))
-        (if (member (ctran-block start)
-                    (gethash path (code-coverage-blocks metadata)))
-            ;; If this source path has already been instrumented in
-            ;; this block, don't instrument it again.
-            start
-            (let ((next (make-ctran))
-                  (*allow-instrumenting* nil))
-              (ensure-gethash path (code-coverage-records metadata)
-                              (cons path +code-coverage-unmarked+))
-              (push (ctran-block start)
-                    (gethash path (code-coverage-blocks metadata)))
-              (ir1-convert start next nil `(%primitive mark-covered ',path))
-              next)))
-      start))
-
-;;; In contexts where we don't have a source location for FORM
-;;; e.g. due to it not being a cons, but where we have a source
-;;; location for the enclosing cons, use the latter source location if
-;;; available. This works pretty well in practice, since many PROGNish
-;;; macroexpansions will just directly splice a block of forms into
-;;; some enclosing form with `(progn ,@body), thus retaining the
-;;; EQness of the conses.
-(defun maybe-instrument-progn-like (start forms form)
-  (or (when (and *allow-instrumenting*
-                 (not (get-source-path form)))
-        (let ((*current-path* (get-source-path forms)))
-          (when *current-path*
-            (instrument-coverage start nil form))))
-      start))
-
 (declaim (start-block find-free-fun find-lexically-apparent-fun
                       ;; needed by ir1-translators
                       find-global-fun))
@@ -271,52 +230,6 @@
                          where
                          (maybe-defined-here name where)))))))
 
-;;; Have some DEFINED-FUN-FUNCTIONALS of a FREE-FUNS entry become invalid?
-;;; Drop 'em.
-;;;
-;;; This was added to fix bug 138 in SBCL. It is possible for a FREE-FUNS
-;;; entry to contain a DEFINED-FUN whose DEFINED-FUN-FUNCTIONAL object
-;;; contained IR1 stuff (NODEs, BLOCKs...) referring to an already compiled
-;;; (aka "dead") component. When this IR1 stuff was reused in a new component,
-;;; under further obscure circumstances it could be used by
-;;; WITH-IR1-ENVIRONMENT-FROM-NODE to generate a binding for
-;;; *CURRENT-COMPONENT*. At that point things got all confused, since IR1
-;;; conversion was sending code to a component which had already been compiled
-;;; and would never be compiled again.
-;;;
-;;; Note: as of 1.0.24.41 this seems to happen only in XC, and the original
-;;; BUGS entry also makes it seem like this might not be an issue at all on
-;;; target.
-(defun clear-invalid-functionals (free-fun)
-  ;; There might be other reasons that FREE-FUN entries could
-  ;; become invalid, but the only one we've been bitten by so far
-  ;; (sbcl-0.pre7.118) is this one:
-  (when (defined-fun-p free-fun)
-    (setf (defined-fun-functionals free-fun)
-          (delete-if (lambda (functional)
-                       (or (eq (functional-kind functional) :deleted)
-                           (when (lambda-p functional)
-                             (or
-                              ;; (The main reason for this first test is to bail
-                              ;; out early in cases where the LAMBDA-COMPONENT
-                              ;; call in the second test would fail because links
-                              ;; it needs are uninitialized or invalid.)
-                              ;;
-                              ;; If the BIND node for this LAMBDA is null, then
-                              ;; according to the slot comments, the LAMBDA has
-                              ;; been deleted or its call has been deleted. In
-                              ;; that case, it seems rather questionable to reuse
-                              ;; it, and certainly it shouldn't be necessary to
-                              ;; reuse it, so we cheerfully declare it invalid.
-                              (not (lambda-bind functional))
-                              ;; If this IR1 stuff belongs to a dead component,
-                              ;; then we can't reuse it without getting into
-                              ;; bizarre confusion.
-                              (eq (component-info (lambda-component functional))
-                                  :dead)))))
-                     (defined-fun-functionals free-fun)))
-    nil))
-
 ;;; If NAME already has a valid entry in (FREE-FUNS *IR1-NAMESPACE*), then return
 ;;; the value. Otherwise, make a new GLOBAL-VAR using information from
 ;;; the global environment and enter it in FREE-FUNS. If NAME
@@ -325,20 +238,7 @@
 ;;; demanded a function.
 (declaim (ftype (sfunction (t string) global-var) find-free-fun))
 (defun find-free-fun (name context &aux (free-funs (free-funs *ir1-namespace*)))
-  (or (let ((old-free-fun (gethash name free-funs)))
-        (when old-free-fun
-          (clear-invalid-functionals old-free-fun)
-          (when (or (not (defined-fun-p old-free-fun))
-                    (not (block-compile *compilation*))
-                    ;; When block-compiling, it is the case that we
-                    ;; could proclaim an inline, define some things,
-                    ;; then proclaim a notinline, and in the case that
-                    ;; the function's :inlinep info changes, the
-                    ;; old-free-fun is stale and we should make a new
-                    ;; one.
-                    (eq (info :function :inlinep name)
-                        (defined-fun-inlinep old-free-fun)))
-            old-free-fun)))
+  (or (gethash name free-funs)
       (let ((kind (info :function :kind name)))
         (ecase kind
           ((:macro :special-form)
@@ -379,16 +279,11 @@
 
 (declaim (end-block))
 
-(defun maybe-find-free-var (name)
-  (let ((found (gethash name (free-vars *ir1-namespace*))))
-    (unless (eq found :deprecated)
-      found)))
-
 ;;; Return the LEAF node for a global variable reference to NAME. If
 ;;; NAME is already entered in (FREE-VARS *IR1-NAMESPACE*), then we just return the
 ;;; corresponding value. Otherwise, we make a new leaf using
 ;;; information from the global environment and enter it in
-;;; FREE-VARS. If the variable is unknown, then we emit a warning.
+;;; FREE-VARS.
 (declaim (ftype (sfunction (t) (or leaf cons heap-alien-info)) find-free-var))
 (defun find-free-var (name &aux (free-vars (free-vars *ir1-namespace*))
                                 (existing (gethash name free-vars)))
@@ -400,8 +295,6 @@
             (type (info :variable :type name))
             (where-from (info :variable :where-from name))
             (deprecation-state (deprecated-thing-p 'variable name)))
-        (when (and (eq kind :unknown) (not deprecation-state))
-          (note-undefined-reference name :variable))
         ;; For deprecated vars, warn about LET and LAMBDA bindings, SETQ, and ref.
         ;; Don't warn again if the name was already seen by the transform
         ;; of SYMBOL[-GLOBAL]-VALUE.
@@ -423,27 +316,53 @@
                        (type (type-specifier (info :variable :type name))))
                    `(macro . (the ,type ,expansion))))
                 (:constant
-                 (find-constant (symbol-value name) name))
+                 (let ((value (symbol-value name)))
+                   ;; Objects that are freely coalescible become
+                   ;; anonymous since we aren't interested in
+                   ;; accessing them through their name.
+                   (if (sb-xc:typep value '(or number character symbol))
+                       (find-constant value)
+                       (make-constant value (ctype-of value) name))))
                 (t
                  (make-global-var :kind kind
                                   :%source-name name
                                   :type type
                                   :where-from where-from)))))))
 
+;;; Return T if and only if OBJ's nature as an externalizable thing renders
+;;; it a leaf for dumping purposes. Symbols are leaflike despite havings slots
+;;; containing pointers; similarly (COMPLEX RATIONAL) and RATIO.
+(defun dumpable-leaflike-p (obj)
+  (or (sb-xc:typep obj '(or symbol number character
+                            ;; (ARRAY NIL) is not included in UNBOXED-ARRAY
+                            (or unboxed-array (array nil))
+                            system-area-pointer
+                            #+sb-simd-pack simd-pack
+                            #+sb-simd-pack-256 simd-pack-256))
+      (cl:typep obj 'debug-name-marker)
+      ;; STANDARD-OBJECT layouts use MAKE-LOAD-FORM, but all other layouts
+      ;; have the same status as symbols - composite objects but leaflike.
+      (and (typep obj 'wrapper) (not (layout-for-pcl-obj-p obj)))
+      ;; PACKAGEs are also leaflike.
+      (cl:typep obj 'package)
+      ;; The cross-compiler wants to dump CTYPE instances as leaves,
+      ;; but CLASSOIDs are excluded since they have a MAKE-LOAD-FORM method.
+      #+sb-xc-host (cl:typep obj '(and ctype (not classoid)))
+      (eq obj sb-lockless:+tail+)))
+
 ;;; Grovel over CONSTANT checking for any sub-parts that need to be
 ;;; processed with MAKE-LOAD-FORM. We have to be careful, because
 ;;; CONSTANT might be circular. We also check that the constant (and
 ;;; any subparts) are dumpable at all.
 (defun maybe-emit-make-load-forms (constant)
   (declare #-sb-xc-host (inline alloc-xset))
-  (dx-let ((xset (alloc-xset)))
+  (dx-let ((things-processed (alloc-xset)))
     (named-let grovel ((value constant))
-      ;; Unless VALUE is an object which which can't contain other objects,
-      ;; or was visited, or is acccessible in a named constant ...
       (unless (or (dumpable-leaflike-p value)
-                  (xset-member-p value xset)
-                  (gethash value (eql-constants *ir1-namespace*)))
-        (add-to-xset value xset)
+                  (xset-member-p value things-processed)
+                  #-sb-xc-host
+                  (unbound-marker-p value))
+        (add-to-xset value things-processed)
         ;; FIXME: shouldn't this be something like SB-XC:TYPECASE ?
         (typecase value
           (cons
@@ -466,11 +385,6 @@
            (dotimes (i (array-total-size value))
              (grovel (row-major-aref value i))))
           (instance
-           ;; In the target SBCL, we can dump any instance, but
-           ;; in the cross-compilation host, %INSTANCE-FOO
-           ;; functions don't work on general instances, only on
-           ;; STRUCTURE!OBJECTs.
-           ;;
            ;; Behold the wonderfully clear sense of this-
            ;;  WHEN (EMIT-MAKE-LOAD-FORM VALUE)
            ;; meaning "when you're _NOT_ using a custom load-form"
@@ -488,25 +402,6 @@
             "Objects of type ~/sb-impl:print-type-specifier/ can't be dumped into fasl files."
             (type-of value)))))))
   (values))
-
-;;; A constant is trivially externalizable if it involves no INSTANCE types
-;;; or any un-dumpable object.
-(defun trivially-externalizable-p (constant)
-  (declare #-sb-xc-host (inline alloc-xset))
-  (dx-let ((xset (alloc-xset)))
-    (named-let ok ((value constant))
-      (if (or (dumpable-leaflike-p value) (xset-member-p value xset))
-          t
-          (progn
-            (add-to-xset value xset)
-            (typecase value
-             (cons (and (ok (car value)) (ok (cdr value))))
-             (vector
-              (dotimes (i (length value) t)
-                (unless (ok (aref value i)) (return nil))))
-             (array
-              (dotimes (i (array-total-size value) t)
-                (unless (ok (row-major-aref value i)) (return nil))))))))))
 
 ;;;; some flow-graph hacking utilities
 
@@ -661,8 +556,12 @@
       (declare (fixnum pos))
       (macrolet ((frob ()
                    `(progn
-                      (when (atom subform) (return))
-                      (let ((fm (car subform)))
+                      (let ((fm (cond ((comma-p subform)
+                                       (comma-expr subform))
+                                      ((atom subform)
+                                       (return))
+                                      (t
+                                       (car subform)))))
                         (when (comma-p fm)
                           (setf fm (comma-expr fm)))
                         (cond ((consp fm)
@@ -680,6 +579,8 @@
                                ;; perfect, but better than nothing.
                                (note-source-path subform pos path)))
                         (incf pos))
+                      (when (comma-p subform)
+                        (return))
                       (setq subform (cdr subform))
                       (when (eq subform trail) (return)))))
         (loop
@@ -724,16 +625,20 @@
   (values))
 
 ;;; Generate a reference to a manifest constant, creating a new leaf
-;;; if necessary.
+;;; if necessary. If we are producing a fasl file, make sure that
+;;; MAKE-LOAD-FORM gets used on any parts of the constant that it
+;;; needs to be.
 (defun reference-constant (start next result value)
   (declare (type ctran start next)
            (type (or lvar null) result))
   (ir1-error-bailout (start next result value)
-                     (let* ((leaf (find-constant value))
-                            (res (make-ref leaf)))
-                       (push res (leaf-refs leaf))
-                       (link-node-to-previous-ctran res start)
-                       (use-continuation res next result)))
+    (when (producing-fasl-file)
+      (maybe-emit-make-load-forms value))
+    (let* ((leaf (find-constant value))
+           (res (make-ref leaf)))
+      (push res (leaf-refs leaf))
+      (link-node-to-previous-ctran res start)
+      (use-continuation res next result)))
   (values))
 
 ;;; Add FUNCTIONAL to the COMPONENT-REANALYZE-FUNCTIONALS, unless it's
@@ -752,18 +657,15 @@
 
 ;;; Generate a REF node for LEAF, frobbing the LEAF structure as
 ;;; needed. If LEAF represents a defined function which has already
-;;; been converted, and is not NOTINLINE, then reference the
-;;; functional instead.
+;;; been converted in the same compilation block, and is not
+;;; NOTINLINE, then reference the functional instead.
 (defun reference-leaf (start next result leaf &optional (name '.anonymous.))
   (declare (type ctran start next) (type (or lvar null) result) (type leaf leaf))
   (assure-leaf-live-p leaf)
   (let* ((type (lexenv-find leaf type-restrictions))
          (leaf (or (and (defined-fun-p leaf)
                         (neq (defined-fun-inlinep leaf) 'notinline)
-                        ;; Only reference defined functionals from named-lambda,
-                        ;; everything else should do this through recognize-known-call,
-                        ;; avoiding copying references to global functions
-                        (defined-fun-named-lambda-p leaf)
+                        (defined-fun-same-block-p leaf)
                         (let ((functional (defined-fun-functional leaf)))
                           (when (and functional (not (functional-kind functional)))
                             (maybe-reanalyze-functional functional))))
@@ -804,45 +706,36 @@
 (defun ir1-convert-var (start next result name)
   (declare (type ctran start next) (type (or lvar null) result) (symbol name))
   (let ((var (or (lexenv-find name vars) (find-free-var name))))
-    (if (and (global-var-p var) (not (always-boundp name)))
-        ;; KLUDGE: If the variable may be unbound, convert using SYMEVAL
-        ;; which is not flushable, so that unbound dead variables signal an
-        ;; error (bug 412, lp#722734): checking for null RESULT is not enough,
-        ;; since variables can become dead due to later optimizations.
-        (ir1-convert start next result
-                     (if (eq (global-var-kind var) :global)
-                         `(sym-global-val ',name)
-                         `(symeval ',name)))
-        (etypecase var
-          (leaf
-           (cond
-             ((lambda-var-p var)
-              (let ((home (ctran-home-lambda-or-null start)))
-                (when home
-                  (sset-adjoin var (lambda-calls-or-closes home))))
-              (when (lambda-var-ignorep var)
-                ;; (ANSI's specification for the IGNORE declaration requires
-                ;; that this be a STYLE-WARNING, not a full WARNING.)
-                #-sb-xc-host
-                (compiler-style-warn "reading an ignored variable: ~S" name)
-                ;; there's no need for us to accept ANSI's lameness when
-                ;; processing our own code, though.
-                #+sb-xc-host
-                (warn "reading an ignored variable: ~S" name))))
-           (reference-leaf start next result var name))
-          ((cons (eql macro)) ; symbol-macro
-           ;; FIXME: the following comment is probably wrong now.
-           ;; If we warn here on :early and :late deprecation
-           ;; then we get an extra warning somehow.
-           ;; This case signals {EARLY,LATE,FINAL}-DEPRECATION-WARNING
-           ;; for symbol-macros. Includes variables, constants,
-           ;; etc. in :FINAL deprecation.
-           (when (eq (deprecated-thing-p 'variable name) :final)
-             (check-deprecated-thing 'variable name))
-           ;; FIXME: [Free] type declarations. -- APD, 2002-01-26
-           (ir1-convert start next result (cdr var)))
-          (heap-alien-info
-           (ir1-convert start next result `(%heap-alien ',var))))))
+    (etypecase var
+      (leaf
+       (when (lambda-var-p var)
+         (let ((home (ctran-home-lambda-or-null start)))
+           (when (and home (neq (lambda-var-home var) home))
+             (sset-adjoin var (lambda-calls-or-closes home))))
+         (when (lambda-var-ignorep var)
+           ;; (ANSI's specification for the IGNORE declaration requires
+           ;; that this be a STYLE-WARNING, not a full WARNING.)
+           #-sb-xc-host
+           (compiler-style-warn "reading an ignored variable: ~S" name)
+           ;; there's no need for us to accept ANSI's lameness when
+           ;; processing our own code, though.
+           #+sb-xc-host
+           (warn "reading an ignored variable: ~S" name)))
+       (maybe-note-undefined-variable-reference var name)
+       (reference-leaf start next result var name))
+      ((cons (eql macro))               ; symbol-macro
+       ;; FIXME: the following comment is probably wrong now.
+       ;; If we warn here on :early and :late deprecation
+       ;; then we get an extra warning somehow.
+       ;; This case signals {EARLY,LATE,FINAL}-DEPRECATION-WARNING
+       ;; for symbol-macros. Includes variables, constants,
+       ;; etc. in :FINAL deprecation.
+       (when (eq (deprecated-thing-p 'variable name) :final)
+         (check-deprecated-thing 'variable name))
+       ;; FIXME: [Free] type declarations. -- APD, 2002-01-26
+       (ir1-convert start next result (cdr var)))
+      (heap-alien-info
+       (ir1-convert start next result `(%heap-alien ',var)))))
   (values))
 
 ;;; Find a compiler-macro for a form, taking FUNCALL into account.
@@ -1050,19 +943,13 @@
 ;;; instrumentation for?
 (defun step-form-p (form)
   (flet ((step-symbol-p (symbol)
-           ;; Consistent treatment of *FOO* vs (SYMBOL-VALUE '*FOO*):
-           ;; we insert calls to SYMEVAL for most non-lexical
-           ;; variable references in order to avoid them being elided
-           ;; if the value is unused.
-           (if (member symbol '(symeval sym-global-val))
-               (not (constantp (second form)))
-               (not (member (sb-xc:symbol-package symbol)
-                            (load-time-value
-                              ;; KLUDGE: packages we're not interested in
-                              ;; stepping.
-                              (mapcar #'find-package '(sb-c sb-int sb-impl
-                                                       sb-kernel sb-pcl))
-                             t))))))
+           (not (member (sb-xc:symbol-package symbol)
+                        (load-time-value
+                         ;; KLUDGE: packages we're not interested in
+                         ;; stepping.
+                         (mapcar #'find-package '(sb-c sb-int sb-impl
+                                                  sb-kernel sb-pcl))
+                         t)))))
     (and *allow-instrumenting*
          (policy *lexenv* (= insert-step-conditions 3))
          (listp form)
@@ -1100,12 +987,16 @@
 ;;; node. FUN-LVAR yields the function to call. ARGS is the list of
 ;;; arguments for the call, which defaults to the cdr of source. We
 ;;; return the COMBINATION node.
-(defun ir1-convert-combination-args (fun-lvar start next result args)
+(defun ir1-convert-combination-args (fun-lvar start next result args
+                                     &key (pass-nargs t)
+                                          arg-source-forms)
   (declare (type ctran start next)
            (type lvar fun-lvar)
            (type (or lvar null) result)
            (type list args))
   (let ((node (make-combination fun-lvar)))
+    (unless pass-nargs
+      (setf (combination-pass-nargs node) nil))
     (setf (lvar-dest fun-lvar) node)
     (collect ((arg-lvars))
       (let ((this-start start)
@@ -1119,7 +1010,9 @@
                  (setf this-start
                        (maybe-instrument-progn-like this-start forms arg))
                  (let ((this-ctran (make-ctran))
-                       (this-lvar (make-lvar node)))
+                       (this-lvar (make-lvar node))
+                       (*current-path* (or (get-source-path (pop arg-source-forms))
+                                           *current-path*)))
                    (ir1-convert this-start this-ctran this-lvar arg)
                    (setq this-start this-ctran)
                    (arg-lvars this-lvar))))
@@ -1182,13 +1075,15 @@
            (type (or lvar null) result)
            (list form)
            (type global-var var))
-  (let ((info (info :function :info (leaf-source-name var))))
-    (if (and info
-             (ir1-attributep (fun-info-attributes info) predicate)
-             (not (if-p (and result (lvar-dest result)))))
-        (let ((*instrument-if-for-code-coverage* nil))
-          (ir1-convert start next result `(if ,form t nil)))
-        (ir1-convert-combination-checking-type start next result form var))))
+  (if (vop-existsp :named sb-vm::move-conditional-result)
+      (ir1-convert-combination-checking-type start next result form var)
+      (let ((info (info :function :info (leaf-source-name var))))
+        (if (and info
+                 (ir1-attributep (fun-info-attributes info) predicate)
+                 (not (if-p (and result (lvar-dest result)))))
+            (let ((*instrument-if-for-code-coverage* nil))
+              (ir1-convert start next result `(if ,form t nil)))
+            (ir1-convert-combination-checking-type start next result form var)))))
 
 ;;; Actually really convert a global function call that we are allowed
 ;;; to early-bind.
@@ -1281,6 +1176,7 @@
                    (var (or bound-var
                             (lexenv-find var-name vars)
                             (find-free-var var-name))))
+              (maybe-note-undefined-variable-reference var var-name)
               (etypecase var
                 (leaf
                  (flet
@@ -1462,8 +1358,10 @@
     (when (defined-fun-p var)
       (setf (defined-fun-inline-expansion res)
             (defined-fun-inline-expansion var))
-      (setf (defined-fun-functionals res)
-            (defined-fun-functionals var)))
+      (setf (defined-fun-functional res)
+            (defined-fun-functional var))
+      (setf (defined-fun-same-block-p res)
+            (defined-fun-same-block-p var)))
     ;; FIXME: Is this really right? Needs we not set the FUNCTIONAL
     ;; to the original global-var?
     res))
@@ -1572,10 +1470,17 @@
          (setf (lambda-var-ignorep var) t)))))
   (values))
 
+(defvar *stack-allocate-dynamic-extent* t
+  "If true (the default), the compiler respects DYNAMIC-EXTENT declarations
+and stack allocates otherwise inaccessible parts of the object whenever
+possible. Potentially long (over one page in size) vectors are, however, not
+stack allocated except in zero SAFETY code, as such a vector could overflow
+the stack without triggering overflow protection.")
+
 (defun process-extent-decl (names vars fvars kind)
   (let ((extent
           (ecase kind
-            (dynamic-extent
+            ((dynamic-extent dynamic-extent-no-note)
              (when *stack-allocate-dynamic-extent*
                kind))
             ((indefinite-extent truly-dynamic-extent)
@@ -1587,22 +1492,21 @@
              (let* ((bound-var (find-in-bindings vars name))
                     (var (or bound-var
                              (lexenv-find name vars)
-                             (maybe-find-free-var name))))
+                             (find-free-var name))))
+               (maybe-note-undefined-variable-reference var name)
                (etypecase var
                  (leaf
-                  (if bound-var
-                      (if (and (leaf-extent var) (neq extent (leaf-extent var)))
-                          (warn "Multiple incompatible extent declarations for ~S?" name)
-                          (setf (leaf-extent var) extent))
-                      (compiler-notify
-                       "Ignoring free ~S declaration: ~S" kind name)))
+                  (cond
+                    ((and (typep var 'global-var) (eq (global-var-kind var) :unknown)))
+                    (bound-var
+                     (if (and (leaf-extent var) (neq extent (leaf-extent var)))
+                         (warn "Multiple incompatible extent declarations for ~S?" name)
+                         (setf (leaf-extent var) extent)))
+                    (t (compiler-notify "Ignoring free ~S declaration: ~S" kind name))))
                  (cons
                   (compiler-error "~S on symbol-macro: ~S" kind name))
                  (heap-alien-info
-                  (compiler-error "~S on alien-variable: ~S" kind name))
-                 (null
-                  (compiler-style-warn
-                   "Unbound variable declared ~S: ~S" kind name)))))
+                  (compiler-error "~S on alien-variable: ~S" kind name)))))
             ((and (consp name)
                   (eq (car name) 'function)
                   (null (cddr name))
@@ -1667,6 +1571,11 @@
         (process-ftype-decl (second spec) res (cddr spec) fvars context))
        ((inline notinline maybe-inline)
         (process-inline-decl spec res fvars))
+       (no-compiler-macro
+        (make-lexenv :default res
+                     :user-data (list*
+                                 (cons 'no-compiler-macro (second spec))
+                                 (lexenv-user-data res))))
        (optimize
         (multiple-value-bind (new-policy specified-qualities)
             (process-optimize-decl spec (lexenv-policy res))
@@ -1682,7 +1591,7 @@
          :default res
          :handled-conditions (process-unmuffle-conditions-decl
                               spec (lexenv-handled-conditions res))))
-       ((dynamic-extent truly-dynamic-extent indefinite-extent)
+       ((dynamic-extent truly-dynamic-extent indefinite-extent dynamic-extent-no-note)
         (process-extent-decl (cdr spec) vars fvars (first spec))
         res)
        ((disable-package-locks enable-package-locks)
@@ -1724,15 +1633,18 @@
      optimize-qualities)))
 
 (defun process-muffle-decls (decls lexenv)
-  (flet ((process-it (spec)
-           (cond ((atom spec))
-                 ((member (car spec) '(muffle-conditions unmuffle-conditions))
-                  (setq lexenv
-                        (process-1-decl spec lexenv nil nil nil nil))))))
-    (dolist (decl decls)
-      (dolist (spec (rest decl))
-        (process-it spec))))
-  lexenv)
+  (let (source-form)
+   (flet ((process-it (spec)
+            (cond ((atom spec))
+                  ((member (car spec) '(muffle-conditions unmuffle-conditions))
+                   (setq lexenv
+                         (process-1-decl spec lexenv nil nil nil nil)))
+                  ((eq (car spec) 'source-form)
+                   (setf source-form (cadr spec))))))
+     (dolist (decl decls)
+       (dolist (spec (rest decl))
+         (process-it spec))))
+    (values lexenv source-form)))
 
 ;;; Use a list of DECLARE forms to annotate the lists of LAMBDA-VAR
 ;;; and FUNCTIONAL structures which are being bound. In addition to
@@ -1802,8 +1714,7 @@
                    ;; E.g. suppressing argument checking without doing
                    ;; so in all the subforms.
                    ((typep spec '(cons (eql local-optimize)))
-                    (setf local-optimize spec)
-                    (setf source-form (cadr spec)))
+                    (setf local-optimize spec))
                    (t
                     (multiple-value-bind (new-env new-qualities)
                         (process-1-decl spec lexenv vars fvars
@@ -1846,7 +1757,7 @@
                             &body forms)
   (declare (symbol ctran lvar))
   (let ((post-binding-lexenv-p (not (null post-binding-lexenv)))
-        (post-binding-lexenv (or post-binding-lexenv (sb-xc:gensym "LEXENV"))))
+        (post-binding-lexenv (or post-binding-lexenv (gensym "LEXENV"))))
     `(%processing-decls ,decls ,vars ,fvars ,ctran ,lvar
                         ,post-binding-lexenv-p
                         (lambda (,ctran ,lvar ,post-binding-lexenv)
@@ -1875,3 +1786,140 @@
                           :where-from :declared))))
 
 (declaim (end-block))
+
+
+;;;; Proclamations:
+;;;
+;;; PROCLAIM changes the global environment, so we must handle its
+;;; compile time side effects if we are to keep the information in the
+;;; (FREE-xxx *IR1-NAMESPACE*) tables up to date. When there is a var
+;;; structure we disown it by replacing it with an updated copy. Uses
+;;; of the variable which were translated before the PROCLAIM will get
+;;; the old version, while subsequent references will get the updated
+;;; information.
+
+;;; If a special block compilation delimiter, then start or end the
+;;; block as appropriate. If :BLOCK-COMPILE is T or NIL, then we
+;;; ignore any start/end block declarations.
+(defun process-block-compile-proclamation (kind entry-points)
+  (if (eq *block-compile-argument* :specified)
+      (ecase kind
+        (start-block
+         (finish-block-compilation)
+         (let ((compilation *compilation*))
+           (setf (block-compile compilation) t)
+           (setf (entry-points compilation) entry-points)))
+        (end-block
+         (finish-block-compilation)))
+      (compiler-notify "ignoring ~S declaration since ~
+                        :BLOCK-COMPILE is not :SPECIFIED"
+                       kind)))
+
+;;; Update function type info cached in (FREE-FUNS *IR1-NAMESPACE*).
+;;; If:
+;;; -- there is a GLOBAL-VAR, then just update the type and remove the
+;;;    name from the list of undefined functions. Someday we should
+;;;    check for incompatible redeclaration.
+;;; -- there is a FUNCTIONAL, then apply the type assertion to that
+;;;    function.  This will only happen during block compilation.
+(defun process-1-ftype-proclamation (name type)
+  (proclaim-as-fun-name name)
+  (let* ((free-funs (free-funs *ir1-namespace*))
+         (var (gethash name free-funs)))
+    (etypecase var
+      (null)
+      (global-var
+       (setf (gethash name free-funs)
+             (let ((kind (global-var-kind var)))
+               (if (defined-fun-p var)
+                   (make-defined-fun
+                    :%source-name name :type type :where-from :declared :kind kind
+                    :inlinep (defined-fun-inlinep var)
+                    :inline-expansion (defined-fun-inline-expansion var)
+                    :same-block-p (defined-fun-same-block-p var)
+                    :functional (defined-fun-functional var))
+                   (make-global-var :%source-name name :type type
+                                    :where-from :declared :kind kind))))
+       (when (defined-fun-p var)
+         (let ((fun (defined-fun-functional var)))
+           (when fun
+             (assert-definition-type fun type
+                                     :unwinnage-fun #'compiler-notify
+                                     :where "this declaration"))))))))
+
+(defun process-ftype-proclamation (spec names)
+  (declare (list names))
+  (let ((type (specifier-type spec)))
+    (unless (csubtypep type (specifier-type 'function))
+      (error "Not a function type: ~/sb-impl:print-type/" spec))
+    (dolist (name names)
+      (process-1-ftype-proclamation name type))))
+
+;;; Replace each old var entry with one having the new type.
+(defun process-type-proclamation (spec names)
+  (declare (list names))
+  (let ((free-vars (free-vars *ir1-namespace*))
+        (type (specifier-type spec)))
+    (dolist (name names)
+      (let ((var (gethash name free-vars)))
+        (etypecase var
+          (null)
+          ;; Constants cannot be redefined, and we already give the
+          ;; constant object the tightest type possible. We will error
+          ;; if the type is incompatible.
+          (constant)
+          (global-var
+           (setf (gethash name free-vars)
+                 (make-global-var :%source-name name
+                                  :type type :where-from :declared
+                                  :kind (global-var-kind var)))))))))
+
+;;; Similar in effect to FTYPE, but change the :INLINEP. Copying the
+;;; global-var ensures that when we substitute a functional for a
+;;; global var (i.e. for DEFUN) that we won't clobber any uses
+;;; declared :NOTINLINE.
+(defun process-inline-proclamation (kind funs)
+  (declare (type (and inlinep (not null)) kind))
+  (dolist (name funs)
+    (proclaim-as-fun-name name)
+    (let* ((free-funs (free-funs *ir1-namespace*))
+           (var (gethash name free-funs)))
+      (etypecase var
+        (null)
+        (global-var
+         (setf (gethash name free-funs)
+               ;; Use the universal type as the local type restriction,
+               ;; since we are processing the proclamation at the top
+               ;; level and hence in a null lexical environment.
+               (make-new-inlinep var kind *universal-type*)))))))
+
+;;; Handle the compile-time side effects of PROCLAIM that don't happen
+;;; at load time. These mostly side-effect the global state of the
+;;; compiler, rather than the global environment.
+(defun %compiler-proclaim (kind args)
+  (case kind
+    ((special global)
+     (dolist (name args)
+       (let* ((free-vars (free-vars *ir1-namespace*))
+              (old (gethash name free-vars)))
+         (when old
+           (ecase (global-var-kind old)
+             ((:special :global))
+             (:unknown
+              (setf (gethash name free-vars)
+                    (make-global-var :%source-name name :type (leaf-type old)
+                                     :where-from (leaf-where-from old)
+                                     :kind (ecase kind
+                                             (special :special)
+                                             (global :global))))))))))
+    ((start-block end-block)
+     #-(and sb-devel sb-xc-host)
+     (process-block-compile-proclamation kind args))
+    (ftype
+     (destructuring-bind (spec &rest args) args
+       (process-ftype-proclamation spec args)))
+    (type
+     (destructuring-bind (spec &rest args) args
+       (process-type-proclamation spec args)))
+    ((inline notinline maybe-inline)
+     (process-inline-proclamation kind args))))
