@@ -44,7 +44,9 @@
   (loop with policy = (lexenv-policy (node-lexenv call))
         for args on (basic-combination-args call)
         for var in (lambda-vars fun)
-        for name = (lambda-var-%source-name var)
+        for name = (or (and (lambda-var-arg-info var )
+                            (arg-info-key (lambda-var-arg-info var )))
+                       (lambda-var-%source-name var))
         do (assert-lvar-type (car args) (leaf-type var) policy
                              (if (eq (functional-kind fun) :optional)
                                  (make-local-call-context fun name)
@@ -72,23 +74,33 @@
   ;; The block may end up being deleted due to cast optimization
   ;; caused by USE-GOOD-FOR-DX-P
   (unless (node-to-be-deleted-p call)
-    (let* ((dx-lvars
+    (let* (no-notes
+           (dx-lvars
              (loop for arg in (basic-combination-args call)
                    for var in (lambda-vars fun)
                    for dx = (leaf-dynamic-extent var)
                    when (and dx arg (not (lvar-dynamic-extent arg)))
-                     collect (cons dx arg))))
+                   collect (cons dx arg)
+                   do
+                   (when (eq dx 'dynamic-extent-no-note)
+                     (setf no-notes dx)))))
       (when dx-lvars
         (let* ((entry (with-ir1-environment-from-node call
                         (make-entry)))
                (cleanup (make-cleanup :kind :dynamic-extent
                                       :mess-up entry
-                                      :nlx-info dx-lvars)))
+                                      :nlx-info dx-lvars
+                                      :dx-kind no-notes)))
           (setf (entry-cleanup entry) cleanup)
           (insert-node-before call entry)
           (setf (node-lexenv call)
                 (make-lexenv :default (node-lexenv call)
                              :cleanup cleanup))
+          (setf (ctran-next (node-prev call)) nil)
+          (let ((ctran (make-ctran)))
+            (with-ir1-environment-from-node call
+              (ir1-convert (node-prev call) ctran nil '(%cleanup-point))
+              (link-node-to-previous-ctran call ctran)))
           ;; Make CALL end its block, so that we have a place to
           ;; insert cleanup code.
           (node-ends-block call)
@@ -300,8 +312,7 @@
                          *lexenv*))
            (xep (ir1-convert-lambda (make-xep-lambda-expression fun)
                                     :debug-name (debug-name
-                                                 'xep (leaf-debug-name fun))
-                                    :system-lambda t)))
+                                                 'xep (leaf-debug-name fun)))))
       (setf (functional-kind xep) :external
             (leaf-ever-used xep) t
             (functional-entry-fun xep) fun
@@ -353,8 +364,7 @@
 ;;; do LET conversion here.
 (defun locall-analyze-fun-1 (fun)
   (declare (type functional fun))
-  (let ((refs (leaf-refs fun))
-        (local-p t))
+  (let ((refs (leaf-refs fun)))
     (dolist (ref refs)
       (let* ((lvar (node-lvar ref))
              (dest (when lvar (lvar-dest lvar))))
@@ -368,12 +378,9 @@
                  (when (eq (functional-kind fun) :deleted)
                    (return-from locall-analyze-fun-1))
                  (unless (eq (basic-combination-kind dest) :local)
-                   (reference-entry-point ref)
-                   (setq local-p nil)))
+                   (reference-entry-point ref)))
                 (t
-                 (reference-entry-point ref)
-                 (setq local-p nil))))))
-    (when local-p (note-local-functional fun)))
+                 (reference-entry-point ref)))))))
 
   (values))
 
@@ -434,7 +441,6 @@
            (clean-component component))))
      (unless did-something
        (return))))
-  (initial-eliminate-dead-code clambdas)
   (values))
 
 ;;; If policy is auspicious and CALL is not in an XEP and we don't seem
@@ -454,7 +460,18 @@
               (with-ir1-environment-from-node call
                 (let* ((*inline-expansions*
                          (register-inline-expansion original-functional call))
-                       (*lexenv* (functional-lexenv original-functional)))
+                       (functional-lexenv (functional-lexenv original-functional))
+                       (call-policy (lexenv-policy (node-lexenv call)))
+                       ;; The inline expansion should be converted
+                       ;; with the policy in effect at this call site,
+                       ;; not the policy saved in the original
+                       ;; functional.
+                       (*lexenv*
+                         (if (eq (lexenv-policy functional-lexenv)
+                                  call-policy)
+                             functional-lexenv
+                             (make-lexenv :default functional-lexenv
+                                          :policy call-policy))))
                   (values nil
                           (ir1-convert-lambda
                            (functional-inline-expansion original-functional)
@@ -701,8 +718,7 @@
                (%funcall ,entry ,@args))
             :debug-name (debug-name 'hairy-function-entry
                                     (lvar-fun-debug-name
-                                     (basic-combination-fun call)))
-            :system-lambda t))))
+                                     (basic-combination-fun call)))))))
     (convert-call ref call new-fun)
     (dolist (ref (leaf-refs entry))
       (convert-call-if-possible ref (lvar-dest (node-lvar ref))))))
@@ -1291,8 +1307,7 @@
 (defun maybe-let-convert (clambda &optional component)
   (declare (type clambda clambda)
            (type (or null component) component))
-  (unless (or (declarations-suppress-let-conversion-p clambda)
-              (functional-has-external-references-p clambda))
+  (unless (declarations-suppress-let-conversion-p clambda)
     ;; We only convert to a LET when the function is a normal local
     ;; function, has no XEP, and is referenced in exactly one local
     ;; call. Conversion is also inhibited if the only reference is in
@@ -1350,7 +1365,7 @@
   (declare (type cblock block1 block2))
   (or (eq block1 block2)
       (let ((cleanup2 (block-start-cleanup block2)))
-        (do-nested-cleanups (cleanup (block-end-lexenv block1) t)
+        (do-nested-cleanups (cleanup block1 t)
           (when (eq cleanup cleanup2)
             (return t))
           (case (cleanup-kind cleanup)
@@ -1420,7 +1435,6 @@
   (declare (type clambda fun))
   (when (and (not (functional-kind fun))
              (not (functional-entry-fun fun))
-             (not (functional-has-external-references-p fun))
              ;; If a functional is explicitly inlined, we don't want
              ;; to assignment convert it, as more call-site
              ;; specialization can be done with inlining.
@@ -1428,6 +1442,7 @@
              (not (block-delete-p (lambda-block fun))))
     (let ((outside-calls nil)
           (outside-calls-ctran nil)
+          (outside-calls-lvar nil)
           (outside-calls-env nil)
           (outside-calls-cleanup nil))
       (when (and (dolist (ref (leaf-refs fun) t)
@@ -1443,6 +1458,9 @@
                            (let ((dest-ctran
                                    (or (node-next dest)
                                        (block-start (first (block-succ (node-block dest))))))
+                                 (dest-lvar
+                                   (and (valued-node-p dest)
+                                        (node-lvar dest)))
                                  (dest-env
                                    (node-home-lambda dest))
                                  (dest-cleanup
@@ -1463,13 +1481,15 @@
                                     ;; cleanups with respect to the
                                     ;; messiest common ancestor.
                                     (unless (and (or (eq (node-derived-type dest) *empty-type*)
-                                                     (and (eq outside-calls-ctran dest-ctran)))
+                                                     (and (eq outside-calls-ctran dest-ctran)
+                                                          (eq outside-calls-lvar dest-lvar)))
                                                  (eq outside-calls-env dest-env)
                                                  (eq outside-calls-cleanup dest-cleanup))
                                       (return nil)))
                                    (t
                                     (setq outside-calls-env dest-env)
                                     (setq outside-calls-ctran dest-ctran)
+                                    (setq outside-calls-lvar dest-lvar)
                                     (setq outside-calls-cleanup dest-cleanup)))
                              (push dest outside-calls))))))
                  (ok-initial-convert-p fun))
