@@ -162,12 +162,13 @@
   (values))
 
 #+call-symbol
-(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node ltn-policy)
-  (declare (ignore ltn-policy))
-  (let ((dest (node-dest node)))
+(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node)
+  (multiple-value-bind (dest dest-lvar)
+      (and (node-lvar node)
+           (principal-lvar-end (node-lvar node)))
     (cond ((and (basic-combination-p dest)
                 (eq (basic-combination-kind dest) :full)
-                (eq (lvar-uses (basic-combination-fun dest)) node)
+                (eq (basic-combination-fun dest) dest-lvar)
                 ;; Everything else can't handle NIL, just don't
                 ;; bother optimizing it.
                 (not (lvar-value-is-nil fun)))
@@ -462,6 +463,7 @@
   (let* ((test (if-test node))
          (use (lvar-uses test)))
     (unless (and (combination-p use)
+                 (immediately-used-p test use)
                  (let ((info (basic-combination-info use)))
                    (and (template-p info)
                         (template-conditional-p info))))
@@ -497,24 +499,17 @@
 ;;; (The default is o.k. for %CATCH, since environment analysis
 ;;; converted the reference to the escape function into a constant
 ;;; reference to the NLX-INFO.)
-(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup)
-                                              node
-                                              ltn-policy)
-  (declare (ignore escape cleanup ltn-policy))
+(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup) node)
   (setf (basic-combination-info node) :funny)
   (setf (node-tail-p node) nil))
 
 ;;; Make sure that arguments of magic functions are not annotated.
 ;;; (Otherwise the compiler may dump its internal structures as
 ;;; constants :-()
-(defoptimizer (%pop-values ltn-annotate) ((%lvar) node ltn-policy)
-  (declare (ignore %lvar node ltn-policy)))
-(defoptimizer (%dummy-dx-alloc ltn-annotate) ((target source) node ltn-policy)
-  (declare (ignore target source node ltn-policy)))
+(defoptimizer (%pop-values ltn-annotate) ((%lvar)))
+(defoptimizer (%dummy-dx-alloc ltn-annotate) ((target source)))
 
-(defoptimizer (%nip-values ltn-annotate) ((&rest lvars)
-                                          node ltn-policy)
-  (declare (ignore node ltn-policy))
+(defoptimizer (%nip-values ltn-annotate) ((&rest lvars))
   ;; Undo the optimization performed by LTN-ANALYZE-MV-CALL,
   ;; which only uses the CSP of the first argument.
   (loop for lvar-lvar in lvars
@@ -553,7 +548,13 @@
                       (funcall (second type) value)
                       (sb-xc:typep value type))))
           (cond (lvar
-                 (and (constant-lvar-p lvar)
+                 (and (if (policy (let ((uses (lvar-uses lvar)))
+                                    (if (consp uses)
+                                        (car uses)
+                                        uses))
+                              (= preserve-constants 3))
+                          (constant-lvar-ignore-types-p lvar nil)
+                          (constant-lvar-p lvar))
                       (type-p (lvar-value lvar) (cdr restr))))
                 (tn
                  (and (eq (tn-kind tn) :constant)
@@ -668,11 +669,12 @@
                        :arg-check
                        :arg-types)))
           ((template-conditional-p template)
-           (let ((dest (lvar-dest lvar)))
-             (if (and (if-p dest)
-                      (immediately-used-p (if-test dest) call))
-                 (values t nil)
-                 (values nil :conditional))))
+           (or (vop-existsp :named sb-vm::move-conditional-result)
+               (let ((dest (lvar-dest lvar)))
+                 (if (and (if-p dest)
+                          (immediately-used-p (if-test dest) call))
+                     (values t nil)
+                     (values nil :conditional)))))
           ((template-results-ok template dtype)
            (values t nil))
           (t
