@@ -711,8 +711,13 @@
 (defstruct (environment (:copier nil))
   ;; the function that allocates this environment
   (lambda (missing-arg) :type clambda :read-only t)
-  ;; a list of all the LAMBDA-VARS and NLX-INFOs needed from enclosing
-  ;; environments by code in this environment.
+  ;; This ultimately converges to a list of all the LAMBDA-VARs and
+  ;; NLX-INFOs needed from enclosing environments by code in this
+  ;; environment. In the meantime, it may be
+  ;;   * NIL at object creation time
+  ;;   * a superset of the correct result, generated somewhat later
+  ;;   * smaller and smaller sets converging to the correct result as
+  ;;     we notice and delete unused elements in the superset
   (closure nil :type list)
   ;; a list of NLX-INFO structures describing all the non-local exits
   ;; into this environment
@@ -1208,10 +1213,11 @@
   (lets nil :type list)
   ;; all the ENTRY nodes in this function and its LETs, or null in a LET
   (entries nil :type list)
-  ;; a set of all the functions directly called from this function
-  ;; (or one of its LETs) using a non-LET local call. This may include
-  ;; deleted functions because nobody bothers to clear them out.
-  (calls (make-sset) :type (or null sset))
+  ;; CLAMBDAs which are locally called by this lambda, and other
+  ;; objects (closed-over LAMBDA-VARs and XEPs) which this lambda
+  ;; depends on in such a way that DFO shouldn't put them in separate
+  ;; components.
+  (calls-or-closes (make-sset) :type (or null sset))
   ;; the TAIL-SET that this LAMBDA is in. This is null during creation.
   ;;
   ;; In CMU CL, and old SBCL, this was also NILed out when LET
