@@ -353,8 +353,6 @@
 
 (defclass standard-reader-method (standard-accessor-method) ())
 (defclass standard-writer-method (standard-accessor-method) ())
-;;; an extension, apparently.
-(defclass standard-boundp-method (standard-accessor-method) ())
 
 ;;; for (SLOT-VALUE X 'FOO) / ACCESSOR-SLOT-VALUE optimization, which
 ;;; can't be STANDARD-READER-METHOD because there is no associated
@@ -362,20 +360,15 @@
 (defclass global-reader-method (accessor-method) ())
 (defclass global-writer-method (accessor-method) ())
 (defclass global-boundp-method (accessor-method) ())
+(defclass global-makunbound-method (accessor-method) ())
 
 (defclass method-combination (metaobject)
   ((%documentation :initform nil :initarg :documentation)))
 
 (defun make-gf-hashset ()
-  ;; Return what is logically a weak hashset, but physically a weak hash-table
-  ;; because we don't implement hashsets.
-  (make-hash-table :test 'eq
-                   :hash-function #'fsc-instance-hash ; stable hash
-                   :weakness :key
+  (make-hashset 64 #'eq #'fsc-instance-hash
+                   :weakness t
                    :synchronized t))
-(defun add-to-weak-hashset (key set) (setf (gethash key set) t))
-(defun remove-from-weak-hashset (key set) (remhash key set))
-(defun weak-hashset-memberp (key set) (gethash key set))
 
 (defclass standard-method-combination (definition-source-mixin
                                        method-combination)
@@ -451,7 +444,11 @@
    (internal-writer-function
      :initform nil
      :initarg :internal-writer-function
-     :accessor slot-definition-internal-writer-function)))
+     :accessor slot-definition-internal-writer-function)
+   (always-bound-p
+     :initform t
+     :initarg :always-bound-p
+     :accessor slot-definition-always-bound-p)))
 
 (defclass direct-slot-definition (slot-definition)
   ((readers
@@ -472,7 +469,7 @@
 (defun uninitialized-accessor-function (type slotd)
   (lambda (&rest args)
     (declare (ignore args))
-    (error "~:(~A~) function~@[ for ~S ~] not yet initialized."
+    (error "~:(~A~) function~@[ for ~S~] not yet initialized."
            type slotd)))
 ;;; We use a structure here, because fast slot-accesses to this information
 ;;; are critical to making SLOT-VALUE-USING-CLASS &co fast: places that need
@@ -482,15 +479,17 @@
             (:copier nil)
             (:constructor make-slot-info
                 (&key slotd typecheck allocation location
-                 (reader (uninitialized-accessor-function :reader slotd))
-                 (writer (uninitialized-accessor-function :writer slotd))
-                 (boundp (uninitialized-accessor-function :boundp slotd)))))
+                   (reader (uninitialized-accessor-function :reader slotd))
+                   (writer (uninitialized-accessor-function :writer slotd))
+                   (boundp (uninitialized-accessor-function :boundp slotd))
+                   (makunbound (uninitialized-accessor-function :makunbound slotd)))))
   (typecheck nil :type (or null function))
   (allocation nil)
   (location nil)
   (reader (missing-arg) :type function)
   (writer (missing-arg) :type function)
-  (boundp (missing-arg) :type function))
+  (boundp (missing-arg) :type function)
+  (makunbound (missing-arg) :type function))
 (declaim (freeze-type slot-info))
 
 (defclass standard-direct-slot-definition (standard-slot-definition
@@ -559,11 +558,6 @@
                            specializer-with-object)
   ((object :initarg :object :reader specializer-object
            :reader eql-specializer-object)
-   ;; created on demand (if and when needed), the CTYPE is the representation
-   ;; of this metaobject as an internalized type object understood by the
-   ;; kernel's type machinery. The CLOS object is really just a MEMBER-TYPE,
-   ;; but the type system doesn't know that.
-   (sb-kernel:ctype)
    ;; Because EQL specializers are interned, any two putative instances
    ;; of EQL-specializer referring to the same object are in fact EQ to
    ;; each other. Therefore a list of direct methods in the specializer can
