@@ -43,9 +43,10 @@
 # include <dlfcn.h>
 #endif
 
-static int decode_locs(lispobj packed_integer, int *offset, int *elsewhere)
+int df_decode_locs(lispobj encoded, int *offset, int *elsewhere)
 {
     struct varint_unpacker unpacker;
+    lispobj packed_integer = listp(encoded) ? CONS(encoded)->cdr : encoded;
     varint_unpacker_init(&unpacker, packed_integer);
     return varint_unpack(&unpacker, offset) && varint_unpack(&unpacker, elsewhere);
 }
@@ -67,14 +68,14 @@ debug_function_from_pc (struct code* code, void *pc)
 
     struct compiled_debug_fun *df = (struct compiled_debug_fun*)native_pointer(di->fun_map);
     int begin, end, elsewhere_begin, elsewhere_end;
-    if (!decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
+    if (!df_decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
         return NULL;
     sword_t offset = (char*)pc - code_text_start(code);
     while (df) {
         struct compiled_debug_fun *next;
         if (df->next != NIL) {
             next = (struct compiled_debug_fun*) native_pointer(df->next);
-            if (!decode_locs(next->encoded_locs, &end, &elsewhere_end))
+            if (!df_decode_locs(next->encoded_locs, &end, &elsewhere_end))
                 return NULL;
         } else {
             next = 0;
@@ -118,6 +119,7 @@ lispobj debug_print(lispobj string)
 {
     print_string(VECTOR(string), stderr);
     putc('\n', stderr);
+    fflush(stderr);
     return 0;
 }
 
@@ -129,7 +131,7 @@ lispobj symbol_package(struct symbol* s)
     // End-users should never see this failure.
     if (!lisp_package_vector) {
         if (!warned) {
-          fprintf(stderr, "Warning: package vector has not been initialized yet\n");
+          fprintf(stderr, "Warning: lisp package array is not initialized for C\n");
           warned = 1;
         }
         return NIL;
@@ -264,13 +266,17 @@ call_info_from_context(struct call_info *info, os_context_t *context)
         info->frame =
             (struct call_frame *)(uword_t)
                 (*os_context_register_addr(context, reg_OCFP));
+#ifdef reg_LRA
         info->lra = (lispobj)(*os_context_register_addr(context, reg_LRA));
+#else
+        info->lra = (lispobj)(*os_context_register_addr(context, reg_RA));
+#endif
         info->code = code_pointer(info->lra);
         pc = (uword_t)native_pointer(info->lra);
     } else
 #endif
     {
-        pc = *os_context_pc_addr(context);
+        pc = os_context_pc(context);
         info->frame =
             (struct call_frame *)(uword_t)
                 (*os_context_register_addr(context, reg_CFP));
@@ -323,12 +329,14 @@ int lisp_frame_previous(struct thread *thread, struct call_info *info)
         info->code =
 #ifdef reg_CODE
         (struct code*)native_pointer(this_frame->code);
-        info->pc = lra;
 #else
         (struct code *)dynamic_space_code_from_pc((char *)lra);
+#endif
+#ifdef reg_LRA
+        info->pc = lra;
+#else
         info->pc = (char*)native_pointer(lra) - (char*)info->code;
 #endif
-
         info->lra = NIL;
     } else {
         info->code = code_pointer(lra);
@@ -365,13 +373,13 @@ lisp_backtrace(int nframes)
         printf("%4d: ", i);
         // Print spaces to keep the alignment nice
         if (info.interrupted
-#ifdef reg_CODE
+#ifdef reg_LRA
             || info.lra == NIL
 #endif
             ) {
             putchar('[');
             if (info.interrupted) { footnotes |= 1; putchar('I'); }
-#ifdef reg_CODE
+#ifdef reg_LRA
             if (info.lra == NIL) { footnotes |= 2; putchar('*'); }
 #endif
             putchar(']');
@@ -394,6 +402,13 @@ lisp_backtrace(int nframes)
             && info.lra != NIL)
             printf("LRA=%p ", (void*)info.lra);
 
+        int fpvalid = (lispobj*)info.frame >= thread->control_stack_start
+          && (lispobj*)info.frame < thread->control_stack_end;
+
+        // If the FP is invalid, then quite likely we'd crash trying to find a
+        // compiled-debug-fun because info.code is a wild pointer
+        if (!fpvalid) { printf(" BAD FRAME\n"); break; }
+
         if (info.code) {
             struct compiled_debug_fun *df;
             if (absolute_pc &&
@@ -410,7 +425,7 @@ lisp_backtrace(int nframes)
 
     } while (++i <= nframes);
     if (footnotes) printf("Note: [I] = interrupted"
-#ifdef reg_CODE
+#ifdef reg_LRA
                           ", [*] = no LRA"
 #endif
                           "\n");
@@ -502,11 +517,13 @@ describe_thread_state(void)
     if (string[0]) printf("Signal mask: %s\n", string);
 #endif
     printf("Specials:\n");
-    printf(" *GC-INHIBIT* = %s\n", (read_TLS(GC_INHIBIT, thread) == T) ? "T" : "NIL");
-    printf(" *GC-PENDING* = %s\n", (read_TLS(GC_PENDING, thread) == T) ? "T" : "NIL");
-    printf(" *INTERRUPTS-ENABLED* = %s\n", (read_TLS(INTERRUPTS_ENABLED, thread) == T) ? "T" : "NIL");
+    printf(" *GC-INHIBIT* = %s\n", read_TLS(GC_INHIBIT, thread) == LISP_T ? "T" : "NIL");
+    printf(" *GC-PENDING* = %s\n", read_TLS(GC_PENDING, thread) == LISP_T ? "T" : "NIL");
+    printf(" *INTERRUPTS-ENABLED* = %s\n",
+           read_TLS(INTERRUPTS_ENABLED, thread) == LISP_T ? "T" : "NIL");
 #ifdef STOP_FOR_GC_PENDING
-    printf(" *STOP-FOR-GC-PENDING* = %s\n", (read_TLS(STOP_FOR_GC_PENDING, thread) == T) ? "T" : "NIL");
+    printf(" *STOP-FOR-GC-PENDING* = %s\n",
+           read_TLS(STOP_FOR_GC_PENDING, thread) == LISP_T ? "T" : "NIL");
 #endif
     printf("Pending handler = %p\n", data->pending_handler);
 }
@@ -572,10 +589,10 @@ void backtrace_from_fp(void *fp, int nframes, int start) {
     log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes, start, stdout);
 }
 
-void backtrace_from_context(os_context_t *context, int nframes) {
+void print_backtrace_from_context(os_context_t *context, int nframes, FILE* file) {
     void *fp = (void *)os_context_frame_pointer(context);
-    print_backtrace_frame((void *)*os_context_pc_addr(context), fp, 0, stdout);
-    backtrace_from_fp(fp, nframes - 1, 1);
+    print_backtrace_frame((void *)os_context_pc(context), fp, 0, file);
+    log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes - 1, 1, file);
 }
 
 void
@@ -586,7 +603,7 @@ lisp_backtrace(int nframes)
 
     if (free_ici) {
         os_context_t *context = nth_interrupt_context(free_ici - 1, thread);
-        backtrace_from_context(context, nframes);
+        print_backtrace_from_context(context, nframes, stdout);
     } else {
         void *fp;
 
@@ -641,7 +658,7 @@ void libunwind_backtrace(struct thread *th, os_context_t *context)
 {
     fprintf(stderr, "Lisp thread @ %p, tid %d", th, (int)th->os_kernel_tid);
 #ifdef LISP_FEATURE_SB_THREAD
-    // the TLS area is not used if #-sb-thread. And if so, it must be "main thred"
+    // the TLS area is not used if #-sb-thread. And if so, it must be "main thread"
     struct thread_instance* lispthread = (void*)native_pointer(th->lisp_thread);
     if (lispthread->name != NIL) {
         fprintf(stderr, " (\"");
@@ -649,6 +666,10 @@ void libunwind_backtrace(struct thread *th, os_context_t *context)
         fprintf(stderr, "\")");
     }
     putc('\n', stderr);
+    // In case you get no backtrace whatsoever, maybe at least see where the
+    // signal was received, probably in a function without the standard
+    // frame pointer setup.
+    fprintf(stderr, " interrupted @ PC %p\n", (void*)OS_CONTEXT_PC(context));
     if (lispthread->waiting_for != NIL) {
         fprintf(stderr, "waiting for %p", (void*)lispthread->waiting_for);
         if (instancep(lispthread->waiting_for)) {
@@ -705,7 +726,7 @@ void backtrace_lisp_threads(int __attribute__((unused)) signal,
     struct thread* this_thread = get_sb_vm_thread();
 #ifdef LISP_FEATURE_SB_THREAD
     if (backtrace_completion_pipe[1] >= 0) {
-        libunwind_backtrace(current_thread, context);
+        libunwind_backtrace(get_sb_vm_thread(), context);
         write(backtrace_completion_pipe[1], context /* any random byte */, 1);
         return;
     }
@@ -744,24 +765,41 @@ static void* watchdog_thread(void* arg) {
     struct timeval timeout;
     fd_set fds;
     FD_ZERO(&fds);
-    FD_SET(watchdog_pipe[0], &fds);
-    timeout.tv_sec = (long)arg;
-    timeout.tv_usec = 0;
-    int nfds = select(watchdog_pipe[0]+1, &fds, 0, 0, &timeout);
-    if (nfds == 0) {
-        // Ensure this message comes out in one piece even if nothing following it does.
-        char msg[] = "Watchdog timer expired\n"; write(2, msg, sizeof msg-1);
-        backtrace_lisp_threads(0, 0, 0);
-        _exit(1); // cause the test suite to exit with failure
+    for (;;) {
+        FD_SET(watchdog_pipe[0], &fds);
+        timeout.tv_sec = (long)arg;
+        timeout.tv_usec = 0;
+        int nfds = select(watchdog_pipe[0]+1, &fds, 0, 0, &timeout);
+        if (nfds < 1) {
+            // Ensure this message comes out in one piece even if nothing following it does.
+            char msg[] = "Watchdog timer expired\n"; write(2, msg, sizeof msg-1);
+            backtrace_lisp_threads(0, 0, 0);
+            _exit(1); // cause the test suite to exit with failure
+        }
+        if (!FD_ISSET(watchdog_pipe[0], &fds))
+            lose("Watchdog got strange file descriptor mask");
+        char byte;
+        int n = read(watchdog_pipe[0], &byte, 1);
+        if (n != 1) lose("watchdog could not read a byte?");
+        fprintf(stderr, "watchdog got byte %d\n", byte);
+        switch (byte) {
+        case 0: return 0;
+        case 1: break; // restart the timer
+        default: lose("watchdog: bad byte");
+        }
     }
     return 0;
 }
-void start_watchdog(int sec) {
+void start_sbcl_watchdog(int sec) {
     if (pipe(watchdog_pipe)) lose("Can't make watchdog pipe");
     pthread_create(&watchdog_tid, 0, watchdog_thread, (void*)(long)sec);
     char msg[] = "Started watchdog thread\n"; write(2, msg, sizeof msg-1);
 }
-void stop_watchdog() {
+void reset_sbcl_watchdog_timer() {
+    char c[1] = {1};
+    write(watchdog_pipe[1], c, 1);
+}
+void stop_sbcl_watchdog() {
     char c[1] = {0};
     write(watchdog_pipe[1], c, 1);
     close(watchdog_pipe[1]);
