@@ -1729,3 +1729,42 @@
                                 m)))
                        (declare (dynamic-extent x))
                        (print x)))))
+
+(with-test (:name :nested-var)
+  (checked-compile-and-assert
+   ()
+   `(lambda ()
+      (multiple-value-bind (v4 c)
+          (values (vector 1) 2)
+        (let ((m v4))
+          (declare (dynamic-extent m))
+          (eval m))
+        (values (sb-ext:heap-allocated-p v4) c)))
+   (() (values :dynamic 2)))
+  (checked-compile-and-assert
+   ()
+   `(lambda ()
+      (flet ((m (&key v4 c)
+               (let ((m v4))
+                 (declare (dynamic-extent m))
+                 (eval m))
+               (values v4 c)))
+        (let ((m (vector 1)))
+          (m :v4 m)
+          (sb-ext:heap-allocated-p m))))
+   (() :dynamic)))
+
+(with-test (:name :dx-do-propagate-let-var)
+  (checked-compile-and-assert
+   ()
+   '(lambda (z)
+     (let ((x (cons nil (let ((y (list z z z)))
+                          (declare (optimize debug))
+                          y))))
+       (declare (dynamic-extent x))
+       (print x)
+       (print x)
+       (assert (sb-ext:stack-allocated-p x))
+       (assert (sb-ext:stack-allocated-p (cdr x))))
+     nil)
+   ((3) NIL)))
