@@ -49,6 +49,8 @@
              ((string= arg "--report-skipped-tests")
               (setf *report-skipped-tests* t))
              ((string= arg "--no-color"))
+             ((string= arg "--slow")
+              (push :slow *features*))
              (t
               (push (merge-pathnames (parse-namestring arg)) *explicit-test-files*))))
   (setf *explicit-test-files* (nreverse *explicit-test-files*))
@@ -200,6 +202,7 @@
       sb-c::*compile-file-elapsed-time*
       sb-impl::*finalizer-rehashlist*
       sb-impl::*finalizers-triggered*
+      sb-impl::*all-packages*
       sb-impl::*package-names-cookie*
       sb-impl::*available-buffers*
       sb-impl::*token-buf-pool*
@@ -528,7 +531,9 @@
      ,*break-on-failure*
      ,*break-on-expected-failure*
      ,*break-on-error*
-     ,(eq *test-evaluator-mode* :interpret))))
+     ,(eq *test-evaluator-mode* :interpret)
+     ,(and (member :slow *features*)
+           t))))
 
 (defun impure-runner (files test-fun log)
   (when files
