@@ -11,6 +11,9 @@
 
 #-x86-64 (invoke-restart 'run-tests::skip-file)
 
+(load "compiler-test-util.lisp")
+(import 'ctu:disassembly-lines)
+
 ;;; This trivial function failed to compile due to rev 88d078fe
 (defun foo (&key k)
   (make-list (reduce #'max (mapcar #'length k))))
@@ -39,22 +42,6 @@
               sb-vm:other-pointer-lowtag)))
     (assert (> a sb-vm:static-space-start))))
 
-(load "compiler-test-util.lisp")
-(defun disassembly-lines (fun)
-  ;; FIXME: I don't remember what this override of the hook is for.
-  (sb-int:encapsulate 'sb-disassem::add-debugging-hooks 'test
-                      (lambda (f &rest args) (declare (ignore f args))))
-  (prog1
-      (mapcar (lambda (x) (string-left-trim " ;" x))
-              (cddr
-               (split-string
-                (with-output-to-string (s)
-                  (let ((sb-disassem:*disassem-location-column-width* 0)
-                        (*print-pretty* nil))
-                    (disassemble fun :stream s)))
-                #\newline)))
-    (sb-int:unencapsulate 'sb-disassem::add-debugging-hooks 'test)))
-
 (sb-vm::define-vop (tryme)
     (:generator 1 (sb-assem:inst mov :byte (sb-vm::ea :gs sb-vm::rax-tn) 0)))
 (with-test (:name :try-gs-segment)
@@ -535,16 +522,20 @@
     ;; Aside from ECASE failure, there are no other JMPs
     (assert (= (count-assembly-labels lines) 1))))
 
+;;; Assert that the ECASE-FAILURE vop emits a trap and that we don't call ERROR
+;;; (cutting down on the code size for each ECASE)
 (with-test (:name :ecase-failure-trap)
-  (assert (null (ctu:find-named-callees
-                 (checked-compile `(lambda (x)
-                                     (ecase x (:a 1) (:b 2) (:c 3)))))))
-  (assert (null (ctu:find-named-callees
-                 (checked-compile `(lambda (x)
+  ;; test ECASE
+  (assert (ctu:asm-search "ECASE-FAILURE-ERROR"
+                                  `(lambda (x)
+                                     (ecase x (:a 1) (:b 2) (:c 3)))))
+  ;; test ETYPECASE
+  (assert (ctu:asm-search "ETYPECASE-FAILURE-ERROR"
+                                  `(lambda (x)
                                      (etypecase x
                                        ((integer 1 20) 'hi)
                                        ((cons (eql :thing)) 'wat)
-                                       (bit-vector 'hi-again))))))))
+                                       (bit-vector 'hi-again))))))
 
 (with-test (:name :symbol-case-optimization-levels)
   (let ((cases
@@ -1193,46 +1184,43 @@
     (split-string (get-output-stream-string string-stream)
                   #\newline)))
 
-(defun assert-has-gc-barrier (match-string lines &optional (skip 0))
-
-  (let (found)
-;; Look for something like this in the trace output:
-;; "VOP CLOSURE-INIT t9[RBX(d)] :NORMAL t13[RSI(d)] :NORMAL {0} "
-;; "        MOV     0, #<TN t14[RAX(u)] :NORMAL>, #<TN t9[RBX(d)] :NORMAL>"
-;; "        SHR     0, #<TN t14[RAX(u)] :NORMAL>, 10"
-;; "        AND     6, #<TN t14[RAX(u)] :NORMAL>, #S(FIXUP :NAME NIL :FLAVOR GC-BARRIER :OFFSET 0)"
-;; "        MOV     4, PTR [R12+RAX+0], 0"
-;; "        MOV     7, PTR [RBX+5], #<TN t13[RSI(d)] :NORMAL>"
-    (loop while lines
-          do
-          (when (and (search match-string (car lines))
-                     (minusp (decf skip)))
-            (setq found t)
-            (let ((barrier-fixup-line (nth 3 lines)))
-              (assert (search ":FLAVOR GC-BARRIER" barrier-fixup-line))
-              (return)))
-          (pop lines))
-    ;; Fail if we didn't find the sentinel CLOSURE-INIT
-    (assert found)))
+;;; Return the list of vops (as strings) that emitted a GC barrier.
+(defun find-gc-barriers (lambda-expression)
+  (let ((lines
+         (compiler-trace-output-lines lambda-expression))
+        (current-vop)
+        (result))
+    ;; FIXME: can this miss vops on a label line, or do we always print labels
+    ;; on their own line?
+    (dolist (line lines (nreverse result))
+     (let ((line (pop lines)))
+       (cond ((and (>= (length line) 4) (string= line "VOP " :end1 4))
+              (let ((string (subseq line 4 (position #\space line :start 5))))
+                (setq current-vop string)))
+             ((search ":FLAVOR GC-BARRIER" line)
+              (push current-vop result)))))))
 
 (with-test (:name :closure-init-gc-barrier)
-  (let ((lines
-         (compiler-trace-output-lines
+  (let ((vops-with-barrier
+         (find-gc-barriers
            '(lambda (address)
-             (declare (sb-vm:word address))
-             (let ((sap (sb-sys:int-sap address)))
-               (lambda () sap))))))
-    (assert-has-gc-barrier "CLOSURE-INIT" lines)))
+              (let ((sap (sb-sys:int-sap (truly-the sb-vm:word address))))
+                ;; The compiler conses the SAP *after* making the closure, so the store
+                ;; of the sap into the closure needs a barrier, as it's an old->young store.
+                (lambda () sap))))))
+    (assert (equal vops-with-barrier '("CLOSURE-INIT")))))
 
 (defstruct (point (:constructor make-point (x))) x (y 0) (z 0))
 (with-test (:name :structure-init-gc-barrier)
-  (let ((lines
-          (compiler-trace-output-lines
+  (let ((vops-with-barrier
+          (find-gc-barriers
            '(lambda (val)
              (declare (double-float val) (inline make-point))
+             ;; Similarly to the closure-init test, DOUBLE-FLOAT reg is cast to
+             ;; DESCRIPTOR-REG only after making the object. This seems like a
+             ;; low-hanging-fruit optimization to flip the order.
              (let ((neg (- val))) (make-point neg))))))
-    (assert-has-gc-barrier "SET-SLOT" lines
-                           #-compact-instance-header 1)))
+    (assert (equal vops-with-barrier '("SET-SLOT")))))
 
 (with-test (:name :system-tlabs)
   (when (find-symbol "SYS-ALLOC-TRAMP" "SB-VM")
@@ -1295,10 +1283,44 @@
                     thereis (search (princ-to-string sym) line))))))
 
 (with-test (:name :closure-gc-barrier)
-  (let ((lines
-          (compiler-trace-output-lines
+  (let ((vops-with-barrier
+         (find-gc-barriers
            '(lambda ()
              (let ((c (cons 1 2)))
                (lambda (x)
                  (setf (car c) x)))))))
-    (assert-has-gc-barrier "SET-SLOT" lines)))
+    (assert (equal vops-with-barrier '("SET-SLOT")))))
+
+(defstruct (wordpair (:predicate nil) (:copier nil)) a b)
+(defun func (x) (wordpair-b x))
+(with-test (:name :non-stack-instance-set)
+  ;; This is a trivial test of the control case for :stack-instance-set
+  (let ((vops-with-barrier
+         (find-gc-barriers
+          '(lambda ()
+            (declare (inline make-wordpair))
+            (let ((pair (make-wordpair :a 'foo)))
+              (setf (wordpair-b pair) "hi")
+              (values (func pair)))))))
+    (assert (equal vops-with-barrier '("INSTANCE-INDEX-SET")))))
+
+(with-test (:name :stack-instance-set)
+  (let ((vops-with-barrier
+         (find-gc-barriers
+          '(lambda ()
+            (declare (inline make-wordpair))
+            (let ((pair (make-wordpair :a 'foo)))
+              (declare (dynamic-extent pair))
+              (setf (wordpair-b pair) "hi")
+              (values (func pair)))))))
+    (assert (not vops-with-barrier))))
+
+(declaim (freeze-type wordpair))
+(with-test (:name :stack-instance-copy)
+  (let ((vops-with-barrier
+         (find-gc-barriers
+          '(lambda (x)
+            (let ((copy (copy-structure (truly-the wordpair x))))
+              (declare (dynamic-extent copy))
+              (values (func copy)))))))
+    (assert (not vops-with-barrier))))
